"""
Module: lfu_cache

This module provides a concrete implementation of a cache using the Least Frequently Used strategy.
The LFUCache class implements the AbstractCache interface to store, retrieve, and manage the states.
It supports caching of Pipeline parameters and global context, along with mechanisms to evict the 
least frequently used items when the cache capacity is exceeded, persist cache state to disk, and 
restore the state from saved checkpoints.
"""

import os
import json
from collections import defaultdict, OrderedDict
import gzip
import dill

from ..utils.utils import create_file, create_directory, get_function_hash
from .abstract_cache import AbstractCache


class LFUCache(AbstractCache):
    """
    LFUCache

    A concrete cache implementation that uses the Least Frequently Used (LFU) strategy.
    Class manages a cache of limited capacity, evicting the least frequently used items when full.
    Also supports persisting cache state to disk (using gzip and dill) and restoring it,
    as well as maintaining a configuration file to track cached steps.
    """

    def __init__(self, capacity=3, cache_dir=None, on_finish='delete'):
        """
        LFUCache Class Constructor method
        Initializes LFUCache Object with necessary parameters:
            - capacity
            - cache_dir
            - on_finish

        Arguments:
            capacity (int): 
                Default: 3
                    Stores a maximum of 3 items in cache at a time.
                int:
                    Maxmimum Number of Items to store in cache at a time.
            cache_dir (string):
                Default: None
                    Cache creates .cache directory using current working directory.
                string:
                    Specific cache directory path can be given.
            on_finish (str: "delete", None):
                Default: "delete"
                    Uses object destructor to delete files and directory when called.
                None:
                    Does nothing when destructor called.
        """
        self.capacity = capacity
        self.min_freq = 0
        self.key_to_val_freq = {}
        self.freq_to_keys = defaultdict(OrderedDict)
        self.on_finish = on_finish

        self.__source_directory = os.path.abspath(os.getcwd())
        self.__cache_directory_path, self.__cache_config_path = self.__init_directory(cache_dir)

    def __del__(self):
        """
        LFUCache Destructor method.
        Deletes necessary components on Program Completion or Object Deletion.
        Uses on_finish property to specify certain actions.
        """
        if self.on_finish == 'delete':
            for file in os.listdir(self.__cache_directory_path):
                file_path = os.path.join(self.__cache_directory_path, file)
                os.remove(file_path)

    def __init_directory(self, cache_dir):
        """
        Public Method: __init_directory()
        Initializes Cache Directory and Configuration File.
        Creates .cache directory in current working directory if cache_dir is None.
        Creates config.json file in cache directory if not present.

        Arguments:
            cache_dir (string):
                Default: None
                    Cache creates .cache directory using current working directory.
                string:
                    Specific cache directory path can be given.

        Returns:
            cache_directory_path (string):
                Path to Cache Directory.
            cache_config_file_path (string):
                Path to Cache Configuration
        """
        if cache_dir is not None:
            cache_directory_path = cache_dir
            cache_config_file_path = None
            for file in os.listdir(cache_directory_path):
                if file.endswith(".json"):
                    cache_config_file_path = os.path.join(cache_directory_path, file)
                    break
            if cache_config_file_path is None:
                cache_config_file_path = os.path.join(cache_directory_path, "config.json")
                create_file(cache_config_file_path)
        else:
            cache_directory_path = os.path.join(self.__source_directory, ".cache")
            create_directory(cache_directory_path)
            cache_config_file_path = os.path.join(cache_directory_path, "config.json")
            create_file(cache_config_file_path)
        return cache_directory_path, cache_config_file_path

    def get(self, key):
        """
        Public Method: get()
        Retrieves value from cache using key.
        Increments Frequency of Key by 1.
        If Key is not present in cache, returns None.

        Arguments:
            key (int):
                Key to retrieve value from cache.

        Returns:
            value (Any):
                Value stored in cache.
            None:
                If Key is not present in cache.
        """
        if key not in self.key_to_val_freq:
            return (None, None)

        value, freq = self.key_to_val_freq[key]
        del self.freq_to_keys[freq][key]
        if not self.freq_to_keys[freq]:
            del self.freq_to_keys[freq]
            if self.min_freq == freq:
                self.min_freq += 1

        new_freq = freq + 1
        self.freq_to_keys[new_freq][key] = None
        self.key_to_val_freq[key] = (value, new_freq)
        return value

    def put(self, value):
        """
        Public Method: put()
        Inserts value into cache.
        If cache is full, evicts least frequently used item.
        If value is a dictionary, extracts parameter_index and globalcontext.

        Arguments:
            value (Any):
                Value to be inserted into cache.

        Returns:
            None:
                If capacity is 0.
        """
        if self.capacity <= 0:
            return

        if isinstance(value, dict) and "parameter_index" in value and "globalcontext" in value:
            value = (value["parameter_index"], value["globalcontext"])

        key = len(self.key_to_val_freq)

        if key in self.key_to_val_freq:
            _, freq = self.key_to_val_freq[key]
            self.key_to_val_freq[key] = (value, freq)
            self.get(key)
            return

        if len(self.key_to_val_freq) >= self.capacity:
            evict_key, _ = self.freq_to_keys[self.min_freq].popitem(last=False)
            if not self.freq_to_keys[self.min_freq]:
                del self.freq_to_keys[self.min_freq]
            del self.key_to_val_freq[evict_key]

        self.key_to_val_freq[key] = (value, 1)
        self.freq_to_keys[1][key] = None
        self.min_freq = 1

    def read_config(self):
        """
        Public Method: read_config()
        Reads Cache Configuration from config.json file.
        If file is not present, returns empty dictionary.

        Returns:
            conf (dict):
                Cache Configuration stored in config.json file.
            {}:
                If config.json file is not present.
        """
        try:
            with open(self.__cache_config_path, 'r', encoding="utf-8") as config_file:
                conf = json.load(config_file)
        except (json.JSONDecodeError, FileNotFoundError):
            conf = {}
        return conf

    def write_config(self, conf):
        """
        Public Method: write_config()
        Writes Cache Configuration to config.json file.
        If file is not present, creates new file.

        Arguments:
            conf (dict):
                Cache Configuration to be stored in config.json file.
        """
        with open(self.__cache_config_path, 'w', encoding="utf-8") as config_file:
            json.dump(conf, config_file, indent=4)

    def delete_cached_file(self, step_key):
        """
        Public Method: delete_cached_file()
        Deletes cached file from cache directory.

        Arguments:
            step_key (string):
                Key of the step to be deleted from cache.
        """
        file_to_delete = os.path.join(self.__cache_directory_path, f"{step_key}.pkl.gz")
        if os.path.exists(file_to_delete):
            os.remove(file_to_delete)

    def update_config(self, step_func, step_key, step_num):
        """
        Public Method: update_config()
        Updates Cache Configuration with new step information.
        If step information is already present, overrides it.
        If step information is not present, adds it to configuration.

        Arguments:
            step_func (function):
                Function to be stored in cache.
            step_key (string):
                Key of the step to be stored in cache.
            step_num (int):
                Index of the step to be stored in cache.
        """
        conf = self.read_config()
        conf['last_completed_step'] = step_key

        if 'steps' not in conf:
            conf['steps'] = OrderedDict()

        source_code, hash_code = get_function_hash(step_func)
        data = {
            "source_code": source_code,
            "func_hash": hash_code
        }
        steps_list = list(conf['steps'].items())

        if step_num < len(steps_list):
            overridden_key, _ = steps_list[step_num]
            if not self.compare_function_code(conf, overridden_key, step_func):
                self.delete_cached_file(overridden_key)
                steps_list.pop(step_num)

        steps_list.insert(step_num, (step_key, data))
        conf['steps'] = OrderedDict(steps_list)

        self.write_config(conf)

    def compare_function_code(self, conf, step_key, func):
        """
        Public Method: compare_function_code()
        Compares function code with configuration function code.
        If function code is same, returns True.
        If function code is different, returns False.

        Arguments:
            conf (dict):
                Cache Configuration to be compared.
            step_key (string):
                Key of the step to be compared.
            func (function):
                Function to be compared.

        Returns:
            True:
                If function code is same.
            False:
                If function code is different.
        """
        current_source_code, current_hash_code = get_function_hash(func)
        conf_source_code = conf['steps'][step_key].get("source_code")
        conf_hash_code = conf['steps'][step_key].get("func_hash")
        if current_hash_code != conf_hash_code:
            return False
        if current_source_code != conf_source_code:
            return False
        return True

    def get_cached_checkpoint(self, step_index):
        """
        Public Method: get_cached_checkpoint()
        Retrieves last completed step from cache configuration.
        Compares function code with configuration function code.
        If function code is different, returns previous step key.
        If function code is same, returns last completed step key.
        If no cached files found, returns None.

        Arguments:
            step_index (OrderedDict):
                Step Index of the Pipeline.

        Returns:
            None:
                If no cached files found.
            previous_step_key (string):
                Key of the previous step to be executed
        """
        conf = self.read_config()
        if conf == OrderedDict():
            # No cached files found
            return None

        last_completed_step = conf['last_completed_step']
        conf_steps = conf['steps']
        previous_step_key = None
        for conf_step_key, pipeline_step_key in zip(conf_steps.keys(), step_index.keys()):
            if conf_step_key != pipeline_step_key:
                return previous_step_key
            if not self.compare_function_code(conf,
                                              conf_step_key,
                                              step_index[conf_step_key].step_func):
                return previous_step_key
            if conf_step_key == last_completed_step:
                break
            previous_step_key = conf_step_key
        return last_completed_step

    def store(self, step_index, parameter_index, global_context, step_key):
        """
        Public Method: store()
        Stores cache state in cache directory.
        Creates a checkpoint file with cache state and step information.

        Arguments:
            step_index (OrderedDict):
                Step Index of the Pipeline.
            parameter_index (dict):
                Parameter Index of the Pipeline.
            global_context (dict):
                Global Context of the Pipeline.
            step_key (string):
                Key of the step to be stored in cache.
        """
        step_num = list(step_index.keys()).index(step_key)
        self.update_config(step_index[step_key].step_func, step_key, step_num)
        checkpoint_file = os.path.join(self.__cache_directory_path, f"{step_key}.pkl.gz")
        cache_state = {
            "capacity": self.capacity,
            "min_freq": self.min_freq,
            "key_to_val_freq": self.key_to_val_freq,
            "freq_to_keys": self.freq_to_keys
        }
        with gzip.open(checkpoint_file, 'wb') as f:
            dill.dump((parameter_index, global_context, cache_state), f)

    def load(self, step_key):
        """
        Public Method: load()
        Loads cache state from cache directory.
        Retrieves cache state and step information from checkpoint file.

        Arguments:
            step_key (string):
                Key of the step to be loaded from cache.

        Returns:
            parameter_index (dict):
                Parameter Index of the Pipeline.
            global_context (dict):
                Global Context of the Pipeline
        """
        checkpoint_file = os.path.join(self.__cache_directory_path, f"{step_key}.pkl.gz")
        with gzip.open(checkpoint_file, 'rb') as f:
            parameter_index, global_context, cache_state = dill.load(f)

        self.capacity = cache_state["capacity"]
        self.min_freq = cache_state["min_freq"]
        self.key_to_val_freq = cache_state["key_to_val_freq"]
        self.freq_to_keys = cache_state["freq_to_keys"]

        return parameter_index, global_context

    def reset(self, delete_directory=False):
        """
        Public Method: reset()
        Resets Cache State.
        Deletes all files in cache directory.
        Deletes cache directory if delete_directory is True.

        Arguments:
            delete_directory (bool):
                Default: False
                    Does not delete cache directory.
                True:
                    Deletes cache directory.
        """
        self.min_freq = 0
        self.key_to_val_freq = {}
        self.freq_to_keys = defaultdict(OrderedDict)

        if delete_directory:
            for file in os.listdir(self.__cache_directory_path):
                file_path = os.path.join(self.__cache_directory_path, file)
                os.remove(file_path)



# """Module: pyodbc_engine.py

# Provides a concrete Engine implementation using pyodbc for connecting to an
# ODBC data source. Defines PyodbcEngine, which manages connection creation,
# context management, and connection testing.
# """

# import pyodbc
# from .engine import Engine

# class PyodbcEngine(Engine):
#     """
#     Concrete Engine for pyodbc-based ODBC connections.

#     Extends the base Engine to create and test a pyodbc.Connection, expose a cursor,
#     and retrieve connection metadata when using a DSN.
#     """

#     def __init__(
#         self,
#         schema: str,
#         dsn: str = None,
#         server: str = "",
#         database: str = "",
#         driver: str = "",
#         **kwargs
#     ):
#         """
#         Initialize a PyodbcEngine.

#         Args:
#             schema (str): Database schema or owner.
#             dsn (str, optional): Data Source Name for ODBC. Defaults to None.
#             server (str): Server hostname or IP address.
#             database (str): Database name.
#             driver (str): ODBC driver name.
#             **kwargs: Additional keyword arguments passed to the base Engine.

#         Raises:
#             RuntimeError: If retrieving connection metadata via DSN fails.
#         """
#         connection_settings = {
#             "server": server,
#             "database": database,
#             "driver": driver,
#             "dsn": dsn
#         }

#         super().__init__(schema, connection_settings, engine_type="pyodbc", **kwargs)
#         self.cursor = self.engine.cursor()

#         if self.dsn is not None:
#             try:
#                 self.driver = self.engine.getinfo(pyodbc.SQL_DRIVER_NAME)
#                 self.database = self.engine.getinfo(pyodbc.SQL_DATABASE_NAME)
#                 self.server = self.engine.getinfo(pyodbc.SQL_SERVER_NAME)
#             except Exception as e:
#                 raise RuntimeError("Error retrieving connection details") from e

#     def create_engine(self):
#         """
#         Create and return a pyodbc.Connection using DSN or explicit connection settings.

#         Returns:
#             pyodbc.Connection: An active ODBC connection with autocommit enabled.

#         Raises:
#             RuntimeError: If establishing the connection fails.
#         """
#         try:
#             if self.dsn:
#                 connection_str = f"DSN={self.dsn};"
#             else:
#                 connection_str = (
#                     f"DRIVER={{{self.driver}}};"
#                     f"SERVER={self.server};"
#                     f"DATABASE={self.database};"
#                     f"SCHEMA={self.schema};"
#                 )
#             return pyodbc.connect(connection_str, autocommit=True)
#         except pyodbc.Error as e:
#             raise RuntimeError("Error establishing connection to the database") from e

#     def test_engine(self):
#         """
#         Validate the pyodbc connection by executing a simple query.

#         Raises:
#             RuntimeError: If the test query execution fails.
#         """
#         try:
#             self.engine.cursor().execute("SELECT 1")
#         except pyodbc.Error as e:
#             raise RuntimeError("Error testing the connection to the database") from e



"""
Module: transform

This module implements a suite of predefined data transformation operations.
It defines a wide range of data transformation tasks including:

    - Base Transformation (Base class for creating predefined transformation steps):
        * Transformation
    - Cache Management (Classes to manage and revert cache states during transformations):
        * CacheState, ReloadCacheState, ResetCache
    - Column Operations (Functions to modify and manage dataframe columns):
        * DropColumn, DropColumns, ConvertColumnType,
          RenameColumns, AddColumn, MergeColumns,
          SplitColumn, ExplodeColumn, CreateColumnFromVariable
    - DataFrame Internal Operations (Functions to manage dataframes within a context):
        * AddDataFrame, DeleteDataFrame, RenameDataFrame, CopyDataFrame
    - Advanced DataFrame Manipulations (Functions for dataframe restructuring and cleaning):
        * TransposeDataFrame, PivotDataFrame, MeltDataFrame,
          GroupByAggregate, FilterRows, SortDataFrame,
          DropDuplicates, SelectColumns, FillNAValues, 
          ReplaceValues, MergeDataFrames, JoinDataFrames,
          ApplyFunction, ApplyMap, MapValues, OneHotEncode
    - Date and Time Conversions (Utility for converting columns to datetime objects):
        * ConvertToDateTime
    - Display Utilities (Functions for summarizing and visualizing dataframe contents):
        * DisplayInfo, DisplayColumns, DisplayHead,
          DisplayTail, DisplayColumnMean, DisplayColumnMedian,
          DisplayColumnMode, DisplayColumnVariance, DisplayColumnStdDev,
          DisplayColumnSum, DisplayColumnMin, DisplayColumnMax,
          DisplayColumnCount, DisplayColumnUnique, DisplayColumnNUnique,
          DisplayColumnDType, DisplayStringCount, DisplayMostFrequentString,
          DisplayAllCategories, DisplaySubstringOccurrence
    - Indexing Operations (Utilities to modify dataframe indices):
        * SetIndex, ResetIndex
    - SQL Integration (Class to execute SQL-based transformations on data):
        * SQLQuery
    - String Manipulations (Functions to handle text cleaning in columns):
        * RemoveCharacterFromColumn, RemoveCharactersFromColumn, ReplaceStringInColumn
    - Variable Operations (Functions to create and manipulate Pipeline variables):
        * CreateVariable, UpdateVariable, DecrementVariable, 
          IncrementVariable, MultiplyVariable, DivideVariable, CopyVariable
    - Aggregation Functions (Statistical functions to aggregate and summarize data):
        * GetColMean, GetColMedian, GetColMode,
          GetColStd, GetColSum, GetColVariance,
          GetColQuantile, GetColCorrelation, GetColCovariance, GetColSkew

The module is designed to be extensible and integrative, allowing for flexible Pipeline
construction and effective data manipulation.
"""
from .transformation import Transformation
from .cache import CacheState
from .cache import ReloadCacheState
from .cache import ResetCache
from .column import DropColumn
from .column import DropColumns
from .column import ConvertColumnType
from .column import RenameColumns
from .column import AddColumn
from .column import MergeColumns
from .column import SplitColumn
from .column import ExplodeColumn
from .column import CreateColumnFromVariable
from .internal import AddDataFrame
from .internal import DeleteDataFrame
from .internal import RenameDataFrame
from .internal import CopyDataFrame
from .dataframe import TransposeDataFrame
from .dataframe import PivotDataFrame
from .dataframe import MeltDataFrame
from .dataframe import GroupByAggregate
from .dataframe import FilterRows
from .dataframe import SortDataFrame
from .dataframe import DropDuplicates
from .dataframe import SelectColumns
from .dataframe import FillNAValues
from .dataframe import ReplaceValues
from .dataframe import MergeDataFrames
from .dataframe import JoinDataFrames
from .dataframe import ApplyFunction
from .dataframe import ApplyMap
from .dataframe import MapValues
from .dataframe import OneHotEncode
from .date import ConvertToDateTime
from .display import DisplayInfo
from .display import DisplayColumns
from .display import DisplayHead
from .display import DisplayTail
from .display import DisplayColumnMean
from .display import DisplayColumnMedian
from .display import DisplayColumnMode
from .display import DisplayColumnVariance
from .display import DisplayColumnStdDev
from .display import DisplayColumnSum
from .display import DisplayColumnMin
from .display import DisplayColumnMax
from .display import DisplayColumnCount
from .display import DisplayColumnUnique
from .display import DisplayColumnNUnique
from .display import DisplayColumnDType
from .display import DisplayStringCount
from .display import DisplayMostFrequentString
from .display import DisplayAllCategories
from .display import DisplaySubstringOccurrence
from .index import SetIndex
from .index import ResetIndex
from .sql import SQLQuery
from .string import RemoveCharacterFromColumn
from .string import RemoveCharactersFromColumn
from .string import ReplaceStringInColumn
from .variable import CreateVariable
from .variable import UpdateVariable
from .variable import DecrementVariable
from .variable import IncrementVariable
from .variable import MultiplyVariable
from .variable import DivideVariable
from .variable import CopyVariable
from .aggregation import GetColMean
from .aggregation import GetColMedian
from .aggregation import GetColMode
from .aggregation import GetColStd
from .aggregation import GetColSum
from .aggregation import GetColVariance
from .aggregation import GetColQuantile
from .aggregation import GetColCorrelation
from .aggregation import GetColCovariance
from .aggregation import GetColSkew



"""
Module: engine

This module implements database engine connectivity.
It provides an interface to interact with SQL-based data sources, 
enabling the execution of queries and the management of database transactions. 
The module currently includes the SQLAlchemyEngine and a PyodbcEngine.
"""
from .engine import AbstractEngine
from .engine import Engine
# from .pyodbc_engine import PyodbcEngine
from .sqlalchemy_engine import SQLAlchemyEngine



"""
Module: csv_extractor

This module provides concrete implementations for extracting data from CSV files.
It defines two classes:
    - CSVExtractor: Extracts a DataFrame from a single CSV file.
    It supports both standard reading and chunked reading when a chunk_size is specified.
    - MultiCSVExtractor: Extracts DataFrames from multiple CSV files located in a directory.
Both classes extend from their respective base extractor classes and leverage pandas'
read_csv() functionality to read data.
Additionally, CSVExtractor provides a method to determine the maximum number of rows
in a CSV file, which is useful for chunking operations.
"""

import pandas as pd
from ..extract.file_extractor import FileExtractor, MultiFileExtractor

class CSVExtractor(FileExtractor):
    """
    CSVExtractor

    A concrete extractor for reading data from CSV files.
    This class extends FileExtractor and provides methods to read an entire CSV file or
    to read it in chunks when a chunk_size is specified. The extracted DataFrame is added
    to the context under the file name.
    """
    def __init__(self,
                 source,
                 step_name="CSVExtractor",
                 chunk_size=None,
                 on_error=None,
                 **kwargs):
        """
        CSVExtractor Class Constructor
        Initializes the CSVExtractor object.

        Arguments:
            source (str): 
                The source directory where the CSV file is located
            step_name (str): 
                The name of the step
            chunk_size (int): 
                The number of rows to read at a time
            on_error (str): 
                The error handling strategy
            **kwargs: 
                Additional keyword arguments for the read_csv() method
        """
        super().__init__(source=source,
                         step_name=step_name,
                         func = self.func,
                         chunk_size=chunk_size,
                         on_error=on_error,
                         **kwargs)

    def func(self, context):
        """
        Public method: func()
        Reads the CSV file and adds the DataFrame to the context

        Arguments:
            context (Context): 
                Blank context object where the DataFrame will be added

        Returns:
            Context: 
                The context object with the DataFrame added
        """
        context.add_dataframe(self.file_name, self.__read_csv(self.file_path, self.kwargs))
        return context

    def __read_csv(self, file, kwargs):
        """
        Private method: __read_csv()
        Reads the CSV file

        Arguments:
            file (str): 
                The path to the CSV file
            kwargs (dict): 
                Additional keyword arguments for the read_csv() method

        Returns:
            DataFrame: 
                The DataFrame read from the CSV file
        """
        return pd.read_csv(file, **kwargs)

    def get_max_row_count(self):
        """
        Public method: get_max_row_count()
        Returns the maximum number of rows in the CSV file

        Returns:
            int: 
                The maximum number of rows in the CSV file
        """
        max_rows = 0
        with open(self.file_path, 'r', encoding="UTF-8") as f:
            row_count = sum(1 for row in f)
            max_rows = max(max_rows, row_count)
        return max_rows

class MultiCSVExtractor(MultiFileExtractor):
    """
    MultiCSVExtractor

    A concrete extractor for reading multiple CSV files from a specified source directory.
    This class extends MultiFileExtractor and leverages CSVExtractor to extract DataFrames
    from each CSV file found in the directory.
    """
    def __init__(self, source, chunk_size=None, on_error=None, **kwargs):
        """
        MultiCSVExtractor Class Constructor
        Initializes the MultiCSVExtractor object.

        Arguments:
            source (str): 
                The source directory where the CSV files are located
            chunk_size (int): 
                The number of rows to read at a time
            on_error (str): 
                The error handling strategy
            **kwargs: 
                Additional keyword arguments for the CSVExtractor constructor
        """
        super().__init__(source=source,
                         step_name="MultiCSVExtractor",
                         type=CSVExtractor,
                         extension_type='csv',
                         chunk_size=chunk_size,
                         on_error=on_error,
                         **kwargs)



"""
Module: string.py

This module implements transformation classes for performing string operations on columns
within a DataFrame. These transformations modify the string values of a specified column
by either removing characters, replacing substrings, or removing multiple characters.
Each transformation extends the base Transformation class and updates the DataFrame in the
Pipeline context accordingly.

Classes:
    RemoveCharacterFromColumn:
                Removes all occurrences of a specified character from a string column.
    RemoveCharactersFromColumn:
                Removes all occurrences of a list of characters from a string column.
    ReplaceStringInColumn:
                Replaces occurrences of a specified substring with another string in a column.
"""

from .transformation import Transformation

class RemoveCharacterFromColumn(Transformation):
    """
    RemoveCharacterFromColumn

    A transformation that removes all occurrences of a specific character from a specified
    string column in a DataFrame. The updated DataFrame is stored back in the context.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        column (str): The name of the column from which to remove the character.
        char_to_remove (str): The character to remove from the column.
    """
    def __init__(self,
                 dataframe,
                 column,
                 char_to_remove,
                 step_name="RemoveCharacterFromColumn",
                 on_error=None):
        """
        Initializes the RemoveCharacterFromColumn transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to be updated in the context.
            column (str): The name of the column to process.
            char_to_remove (str): The character to remove from the column's string values.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "RemoveCharacterFromColumn".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.column = column
        self.char_to_remove = char_to_remove
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the RemoveCharacterFromColumn transformation.

        Retrieves the DataFrame from the context, removes the specified character from
        the target column, updates the DataFrame in the context, and returns the updated context.

        Arguments:
            context (Context): The Pipeline context containing the DataFrame.

        Returns:
            Context: The updated context with the modified DataFrame.
        """
        df = context.dataframes[self.dataframe]
        df[self.column] = self.__remove_char(df)
        context.set_dataframe(self.dataframe, df)
        return context

    def __remove_char(self, df):
        """
        Removes all occurrences of the specified character from the target column.

        Arguments:
            df (DataFrame): The DataFrame to process.

        Returns:
            Series: The modified column with the character removed.
        """
        return df[self.column].str.replace(self.char_to_remove, "", regex=False)


class RemoveCharactersFromColumn(Transformation):
    """
    RemoveCharactersFromColumn

    A transformation that removes all occurrences of a list of specified characters from
    a target string column in a DataFrame. The updated DataFrame is stored back in the context.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        column (str): The name of the column to process.
        chars_to_remove (iterable): An iterable of characters to remove from the column.
    """
    def __init__(self,
                 dataframe,
                 column,
                 chars_to_remove,
                 step_name="RemoveCharactersFromColumn",
                 on_error=None):
        """
        Initializes the RemoveCharactersFromColumn transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update in the context.
            column (str): The name of the column to process.
            chars_to_remove (iterable): A list or iterable of characters to remove from
                                        the column's string values.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "RemoveCharactersFromColumn".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.column = column
        self.chars_to_remove = chars_to_remove  # list or iterable of characters
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the RemoveCharactersFromColumn transformation.

        Retrieves the DataFrame from the context, removes all specified characters from
        the target column, updates the DataFrame in the context, and returns the updated
        context.

        Arguments:
            context (Context): The Pipeline context containing the DataFrame.

        Returns:
            Context: The updated context with the modified DataFrame.
        """
        df = context.dataframes[self.dataframe]
        context.set_dataframe(self.dataframe, self.__remove_chars(df))
        return context

    def __remove_chars(self, df):
        """
        Iteratively removes all specified characters from the target column.

        Arguments:
            df (DataFrame): The DataFrame to process.

        Returns:
            DataFrame: The DataFrame with the specified characters removed from the target column.
        """
        for char in self.chars_to_remove:
            df[self.column] = df[self.column].str.replace(char, "", regex=False)
        return df


class ReplaceStringInColumn(Transformation):
    """
    ReplaceStringInColumn

    A transformation that replaces all occurrences of a specified substring with another
    string in a target column of a DataFrame. The updated DataFrame is stored back in
    the context.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        column (str): The name of the column to process.
        to_replace (str): The substring to be replaced.
        replacement (str): The string to replace with.
    """
    def __init__(self,
                 dataframe,
                 column,
                 to_replace,
                 replacement,
                 step_name="ReplaceStringInColumn",
                 on_error=None):
        """
        Initializes the ReplaceStringInColumn transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update in the context.
            column (str): The name of the column in which the replacement is to occur.
            to_replace (str): The substring to be replaced.
            replacement (str): The string to replace the substring with.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "ReplaceStringInColumn".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.column = column
        self.to_replace = to_replace
        self.replacement = replacement
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the ReplaceStringInColumn transformation.

        Retrieves the DataFrame from the context, replaces occurrences of the specified
        substring in the target column, updates the DataFrame in the context, and
        returns the updated context.

        Arguments:
            context (Context): The Pipeline context containing the DataFrame.

        Returns:
            Context: The updated context with the modified DataFrame.
        """
        df = context.dataframes[self.dataframe]
        df[self.column] = self.__replace_string(df)
        context.set_dataframe(self.dataframe, df)
        return context

    def __replace_string(self, df):
        """
        Replaces all occurrences of the specified substring in the target column with the
        replacement string.

        Arguments:
            df (DataFrame): The DataFrame to process.

        Returns:
            Series: The modified column with the substring replaced.
        """
        return df[self.column].str.replace(self.to_replace, self.replacement, regex=False)



"""
Module: extract

This module implements the extractor functionalities. It provides a set of 
extractor classes that are designed to retrieve data from various sources. 
These include the base Extractor class and specialized extractors for 
handling file-based data, CSV files, Excel files, and SQL Server databases. 
The module also supports handling multiple extractors simultaneously through the MultiExtractor 
variants.
Any custom Extractor class should inherit from the base Extractor class and implement the
required extraction logic for the specific data source.

Key Components:
    - Extractor: Base class defining the common interface for data extraction.
    - MultiExtractor: Supports combining multiple extractor instances.
    - FileExtractor & MultiFileExtractor: Extractors for handling generic file-based data sources.
    - CSVExtractor & MultiCSVExtractor: Extractors for retrieving data from CSV files.
    - ExcelExtractor & MultiExcelExtractor: Extractors for retrieving data from Excel files.
    - SQLServerExtractor: Extractor for connecting to data from SQL Server databases.

These extractor classes facilitate robust and flexible data ingestion into the Pipeline,
ensuring seamless integration with various data sources.
"""

from .extractor import Extractor
from .extractor import MultiExtractor
from .file_extractor import FileExtractor
from .file_extractor import MultiFileExtractor
from .csv_extractor import CSVExtractor
from .csv_extractor import MultiCSVExtractor
from .excel_extractor import ExcelExtractor
from .excel_extractor import MultiExcelExtractor
from .sqlserver_extractor import SQLServerExtractor
from .odbc_extractor import ODBCExtractor



"""
Module: sql.py

This module provides a transformation class for executing SQL queries using the pandasql
library on DataFrames. The SQLQuery transformation enables users to run SQL commands
against the dataframes by referencing them by their keys within the context, and it stores
the query result in the context under a specified name.
"""

#store result in variable

import pandasql as sqldf
from .transformation import Transformation

class SQLQuery(Transformation):
    """
    SQLQuery Transformation

    Executes a SQL query using the pandasql library on DataFrames contained in the Pipeline
    context. The SQL query can reference any DataFrame in the context by its key. The
    result of the query is stored in the context under the provided output DataFrame name.
    """
    def __init__(self,
                 query,
                 output_dataframe_name,
                 step_name="SQLQuery",
                 on_error=None):
        """
        Initializes the SQLQuery transformation.

        Arguments:
            query (str): The SQL query to be executed. The query can reference DataFrames
                         in the context by their keys.
            output_dataframe_name (str): The key under which the query result will be
                                         stored in the context.
            step_name (str, optional): The name of this transformation step.
                                       Defaults to "SQLQuery".
            on_error (str, optional): The error handling strategy.
        """
        self.query = query
        self.output_dataframe_name = output_dataframe_name
        super().__init__(step_name=step_name,
                         func=self.func,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the SQL query on the DataFrames in the context.

        Uses the pandasql library to execute the provided SQL query, treating the context's
        dataframes as the namespace. The resulting DataFrame is then added to the context
        under the specified output name.

        Arguments:
            context (Context): The context object containing the DataFrames.

        Returns:
            Context: The updated context with the SQL query result added.
        """
        result = sqldf.sqldf(self.query, context.dataframes)
        context.set_dataframe(self.output_dataframe_name, result)
        return context



"""
Module: internal.py

This module implements transformation classes for managing DataFrames within the Pipeline context.
It provides functionality for adding, deleting, renaming, and copying DataFrames in the context.
These operations help maintain and modify the state of DataFrames as the Pipeline executes,
ensuring that the context reflects the desired DataFrame structure for subsequent processing steps.

Classes:
    AddDataFrame: Adds a new DataFrame to the context.
    DeleteDataFrame: Deletes a DataFrame from the context.
    RenameDataFrame: Renames an existing DataFrame in the context.
    CopyDataFrame: Creates a copy of an existing DataFrame under a new name.
"""

from .transformation import Transformation

class AddDataFrame(Transformation):
    """
    AddDataFrame

    A transformation that adds a new DataFrame to the Pipeline context.
    The DataFrame is stored in the context under the specified name.
    """
    def __init__(self,
                 dataframe,
                 name,
                 step_name="AddDataFrame",
                 on_error=None):
        """
        Initializes the AddDataFrame transformation.

        Arguments:
            dataframe (DataFrame): The DataFrame to be added.
            name (str): The name under which the DataFrame will be stored in the context.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "AddDataFrame".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.name = name
        super().__init__(step_name=step_name,
                         func=self.func,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the AddDataFrame transformation.

        Adds the specified DataFrame to the context under the provided name.

        Arguments:
            context (Context): The context object where the DataFrame will be added.

        Returns:
            Context: The updated context with the new DataFrame.
        """
        context.add_dataframe(self.name, self.dataframe)
        return context

class DeleteDataFrame(Transformation):
    """
    DeleteDataFrame

    A transformation that deletes a DataFrame from the Pipeline context.
    """
    def __init__(self,
                 dataframe,
                 step_name="DeleteDataFrame",
                 on_error=None):
        """
        Initializes the DeleteDataFrame transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to be deleted from the context.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "DeleteDataFrame".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the DeleteDataFrame transformation.

        Deletes the specified DataFrame from the context.

        Arguments:
            context (Context): The context object containing the DataFrame.

        Returns:
            Context: The updated context with the DataFrame removed.
        """
        del context.dataframes[self.dataframe]
        return context

class RenameDataFrame(Transformation):
    """
    RenameDataFrame

    A transformation that renames a DataFrame within the Pipeline context.
    """
    def __init__(self,
                 old_name,
                 new_name,
                 step_name="RenameDataFrame",
                 on_error=None):
        """
        Initializes the RenameDataFrame transformation.

        Arguments:
            old_name (str): The current name of the DataFrame in the context.
            new_name (str): The new name to assign to the DataFrame.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "RenameDataFrame".
            on_error (str, optional): The error handling strategy.
        """
        self.old_name = old_name
        self.new_name = new_name
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=old_name,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the RenameDataFrame transformation.

        Renames the specified DataFrame in the context by removing it under the old
        name and re-adding it under the new name.

        Arguments:
            context (Context): The context containing the DataFrame.

        Returns:
            Context: The updated context with the DataFrame renamed.
        """
        if self.old_name in context.dataframes:
            df = context.dataframes.pop(self.old_name)
            context.set_dataframe(self.new_name, df)
        return context

class CopyDataFrame(Transformation):
    """
    CopyDataFrame

    A transformation that creates a copy of an existing DataFrame under a new name in
    the Pipeline context.
    """
    def __init__(self,
                 source_dataframe,
                 target_dataframe,
                 step_name="CopyDataFrame",
                 on_error=None):
        """
        Initializes the CopyDataFrame transformation.

        Arguments:
            source_dataframe (str): The name of the existing DataFrame to copy.
            target_dataframe (str): The name under which the copy will be stored in the context.
            step_name (str, optional): The name of this transformation step.
                                       Defaults to "CopyDataFrame".
            on_error (str, optional): The error handling strategy.
        """
        self.source_dataframe = source_dataframe
        self.target_dataframe = target_dataframe
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=source_dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the CopyDataFrame transformation.

        Retrieves the source DataFrame from the context, creates a copy of it,
        stores the copy under the target name, and returns the updated context.

        Arguments:
            context (Context): The context containing the DataFrame.

        Returns:
            Context: The updated context with the new copy of the DataFrame.
        """
        df = context.dataframes[self.source_dataframe].copy()
        context.set_dataframe(self.target_dataframe, df)
        return context



"""
Module: cache.py

This module implements transformation steps related to caching operations.
It provides three transformation classes that interact with a caching mechanism to store, 
reload, or reset states. These transformations allow the execution to branch and
resume from a saved state, or clear cached data as needed.

Classes:
    CacheState: Caches the current state of execution (i.e., parameter index and global context) 
                by storing a deep copy in the cache.
    ReloadCacheState: Reloads a cached state from the cache using a specified cache key and
                updates the Pipeline's parameter index and global context.
    ResetCache: Resets the cache by clearing all cached data, with an option to delete the
                underlying cache directory.
"""

from copy import deepcopy
from .transformation import Transformation

class CacheState(Transformation):
    """
    CacheState Transformation

    This transformation caches the current state of Pipeline execution by storing deep copies
    of the parameter index and global context in the provided cache. 
    It is typically used to create a checkpoint in the Pipeline, allowing execution to resume
    from this point later if necessary.

    Attributes:
        cache: The cache object where the state is stored.
        parameter_index (dict): The current parameter index of the Pipeline.
        globalcontext: The current global context of the Pipeline.
    """
    def __init__(self,
                 cache,
                 parameter_index,
                 globalcontext,
                 step_name="cache_state",
                 on_error=None):
        """
        Initializes the CacheState transformation.

        Arguments:
            cache: The cache object used for storing the Pipeline state.
            parameter_index (dict): The parameter index to be cached.
            globalcontext: The global context (containing dataframes) to be cached.
            step_name (str, optional): The name of this transformation step.
                                       Defaults to "cache_state".
            on_error (str, optional): The error handling strategy.
        """
        super().__init__(step_name=step_name,
                         func=self.func,
                         on_error=on_error)
        self.cache = cache
        self.parameter_index = parameter_index
        self.globalcontext = globalcontext

    def start_step(self):
        """
        Prepares the CacheState transformation for execution.

        This implementation does not require any pre-execution steps.
        """
        return

    def stop_step(self):
        """
        Cleans up after executing the CacheState transformation.

        No cleanup is required for this transformation.
        """
        return

    def func(self):
        """
        Executes the CacheState transformation.

        Creates deep copies of the current parameter index and global context, and stores them
        in the cache using the cache's put() method.
        """
        data = {
            "parameter_index": deepcopy(self.parameter_index),
            "globalcontext": deepcopy(self.globalcontext)
        }
        self.cache.put(data)

    def __str__(self):
        """
        Returns a string representation of the cache used by this transformation.

        Returns:
            str: A string representation of the cache object.
        """
        return f"{self.cache}"


class ReloadCacheState(Transformation):
    """
    ReloadCacheState Transformation

    This transformation reloads a previously cached state from the cache using a specified
    cache key. The reloaded state (parameter index and global context) is then applied to
    the Pipeline, allowing execution to resume from a saved checkpoint.
    """
    def __init__(self,
                 cache_key,
                 cache,
                 pipeline,
                 step_name="reload_cached_state",
                 on_error=None):
        """
        Initializes the ReloadCacheState transformation.

        Arguments:
            cache_key: The key corresponding to the cached state to reload.
            cache: The cache object from which the state is retrieved.
            pipeline: The Pipeline object whose state will be updated.
            step_name (str, optional): The name of this transformation step.
                                       Defaults to "reload_cached_state".
            on_error (str, optional): The error handling strategy.
        """
        super().__init__(step_name=step_name,
                         func=self.func,
                         on_error=on_error)
        self.cache_key = cache_key
        self.cache = cache
        self.pipeline = pipeline

    def start_step(self):
        """
        Prepares the ReloadCacheState transformation for execution.

        No pre-execution steps are required.
        """
        return

    def stop_step(self):
        """
        Cleans up after executing the ReloadCacheState transformation.

        No cleanup is required.
        """
        return

    def func(self):
        """
        Executes the ReloadCacheState transformation.

        Retrieves the cached state using the cache key and updates the Pipeline's
        parameter index and global context accordingly.
        """
        parameter_index, globalcontext = self.cache.get(self.cache_key)
        self.pipeline.parameter_index = parameter_index
        self.pipeline.globalcontext = globalcontext

    def __str__(self):
        """
        Returns a string representation of the cache used by this transformation.

        Returns:
            str: A string representation of the cache object.
        """
        return f"{self.cache}"


class ResetCache(Transformation):
    """
    ResetCache Transformation

    This transformation resets the cache by clearing all cached state. It can optionally
    delete the underlying cache directory, ensuring that all cached data is removed.
    """
    def __init__(self,
                 cache,
                 step_name="reset_cache",
                 delete_directory=False,
                 on_error=None):
        """
        Initializes the ResetCache transformation.

        Arguments:
            cache: The cache object to be reset.
            step_name (str, optional): The name of this transformation step.
                                       Defaults to "reset_cache".
            delete_directory (bool, optional): If True, deletes the cache directory.
                                       Defaults to False.
            on_error (str, optional): The error handling strategy.
        """
        super().__init__(step_name=step_name,
                         func=self.func,
                         on_error=on_error)
        self.cache = cache
        self.delete_directory = delete_directory

    def start_step(self):
        """
        Prepares the ResetCache transformation for execution.

        No pre-execution initialization is needed.
        """
        return

    def stop_step(self):
        """
        Cleans up after executing the ResetCache transformation.

        No cleanup is required for this transformation.
        """
        return

    def func(self):
        """
        Executes the ResetCache transformation.

        Calls the reset() method on the cache object,
        optionally deleting the cache directory if specified.
        """
        self.cache.reset(delete_directory=self.delete_directory)

    def __str__(self):
        """
        Returns a string representation of the cache used by this transformation.

        Returns:
            str: A string representation of the cache object.
        """
        return f"{self.cache}"



"""
Module: abstract_cache

This module defines the AbstractCache class, an abstract base class that specifies the
required interface for caching mechanisms used within the Pipeline framework.
Any concrete cache implementation must inherit from AbstractCache and implement the
methods for inserting, retrieving, storing, loading, and resetting cache items.
This ensures consistent cache behavior across different caching strategies.
"""

from abc import ABC, abstractmethod

class AbstractCache(ABC):
    """
    AbstractCache

    An abstract base class for implementing caching mechanisms within the Pipeline framework.
    Derived classes must implement the following methods to handle caching operations:
      - put: Insert or update an item in the cache.
      - get: Retrieve an item from the cache using a key.
      - store: Cache the current state of the Pipeline.
      - load: Reload a cached state using a specific step key.
      - reset: Reset the cache, with the option to delete the underlying cache directory.
    """

    @abstractmethod
    def put(self, value):
        """
        Abstract Method: put()
        Ensures inherited class instantiates a put method.
        Used to Cache items in Execution.
        Put method inserts/updates a cached item.

        Arguments:
            value (Any):
                Value to cache.
        """

    @abstractmethod
    def get(self, key):
        """
        Abstract Method: get()
        Ensures inherited class instantiates a get method.
        Used to retrieve items in Execution.
        Get method retrieves a cached item using the corresponding key.

        Arguments:
            key (string):
                Key for cached state to retrieve.
        """

    @abstractmethod
    def store(self, step_index, parameter_index, global_context, step_key):
        """
        Abstract Method: store()
        Ensures inherited class instantiates a store method.
        Store method caches the current state including any properties or items desired.

        Arguments:
            step_index (OrderedDict):
                Index containing all steps instantiated in Pipeline object.
            parameter_index (dict):
                Current state of parameter index.
            global_context (Context):
                Current state of Pipeline global context.
            step_key (string):
                Step Key for step to cache.
        """

    @abstractmethod
    def load(self, step_key):
        """
        Abstract Method: load()
        Ensures inherited class instantiates a load method.
        Load method reloades a cached state using the provided step key.

        Arguments:
            step_key (string):
                Step Key for cached state to reload.
        """

    @abstractmethod
    def reset(self, delete_directory=False):
        """
        Abstract Method: reset()
        Ensures inherited class instantiates a reset method.
        Used to reset cache to desired state.

        Arguments:
            delete_directory (Bool):
                True: .cache directory is deleted on reset
                False: .cache directory is not deleted on reset
        """



"""
Module: load

This module provides the loader functionalities. It defines classes responsible
for loading data into target destinations from various sources. 
The module includes a base Loader class along with several specialized loaders 
for different file formats and database systems, such as CSV, Excel, and SQL Server.
Any custom Loader class should inherit from the base Loader class and implement the
required loading logic for the specific data source.

Key Components:
    - Loader: The base loader class that defines the common interface for all data loaders.
    - FileLoader: Implements loading functionality for generic file-based data sources.
    - CSVLoader & MultiCSVLoader: Specialized loaders for handling CSV files.
    - ExcelLoader & MultiExcelLoader: Specialized loaders for handling Excel files.
    - SQLServerLoader: Loader implementation for loading data into SQL Server databases.
"""
from .loader import Loader
from .file_loader import FileLoader
from .csv_loader import CSVLoader
from .csv_loader import MultiCSVLoader
from .excel_loader import ExcelLoader
from .excel_loader import MultiExcelLoader
from .sqlserver_loader import SQLServerLoader
from .odbc_loader import ODBCLoader



"""
Module: transformation.py

This module implements the base Transformation class, which serves as an abstract
extension of the Step class. Transformation provides a common structure for all
transformation operations in the ETL Pipeline. These operations typically modify or
analyze DataFrames and update the Pipeline context accordingly. Subclasses must implement
the abstract method func() to define their specific transformation logic.
"""

from abc import abstractmethod
from ..step.step import Step

class Transformation(Step):
    """
    Transformation Class

    The Transformation class is an abstract base class for defining data transformations
    within the ETL Pipeline. It extends the Step class and provides a common framework
    for executing transformations on DataFrames. Subclasses must implement the abstract
    func() method to provide the actual transformation logic.
    """

    def __init__(self,
                 step_name,
                 func,
                 dataframes=None,
                 on_error=None):
        """
        Initializes the Transformation object.

        Arguments:
            step_name (str): The name of the transformation step.
            func (function): The function that implements the transformation logic.
            dataframes (str or list, optional): The name(s) of the DataFrame(s) in the
                                                context that the transformation will operate on.
                                                If a single DataFrame name is provided,
                                                it is converted to a list.
            on_error (str, optional): The error handling strategy.
        """
        super().__init__(step_name=step_name,
                         func=func,
                         dataframes=dataframes if isinstance(dataframes, list) else [dataframes],
                         on_error=on_error)

    def start_step(self):
        """
        Prepares the transformation for execution.

        This method is called before the transformation function (func) is executed.
        In this base class, no initialization is required, so it simply returns.
        """
        return

    def stop_step(self):
        """
        Cleans up after the transformation execution.

        Clears the parameters dictionary to reset the transformation state,
        ensuring the step can be executed again if necessary.
        """
        self.params.clear()

    @abstractmethod
    def func(self, context):
        """
        Abstract method: func()

        Defines the core transformation logic to be applied to the DataFrame(s) in the context.
        Subclasses must override this method to implement specific transformations.

        Arguments:
            context (Context): The Pipeline context containing the DataFrame(s) to be transformed.

        Returns:
            Context: The updated context after applying the transformation.
        """



"""Module: odbc_loader.py

This module provides a concrete implementation for loading pandas DataFrames into
an ODBC-accessible database (e.g., SQL Server) using raw SQL executed via a pyodbc.Connection.
It defines the ODBCLoader class which writes a DataFrame to a target table, supporting append,
replace, or error-if-exists modes, and automatic table creation with datatype inference.
"""

import pandas as pd
from ..load.loader import Loader

class ODBCLoader(Loader):
    """
    Loader for ODBC-accessible targets. Inserts a pandas DataFrame into a database table via ODBC.

    Supports three existence behaviors:
        - append (default): Add rows to an existing table.
        - replace: Drop and recreate the table before loading.
        - error: Raise if the target table already exists.
    """

    def __init__(
        self,
        target: str,
        engine,
        dataframe: str,
        schema: str,
        step_name: str = "ODBCLoader",
        exists: str = "append",
        on_error=None,
        **kwargs
    ):
        """
        Initialize an ODBCLoader.

        Args:
            target (str): Name of the destination table.
            engine (pyodbc.Connection): Active ODBC connection.
            dataframe (str): Name of DataFrame to load.
            schema (str): Database schema or owner for qualifying the target.
            step_name (str): Identifier for this load step.
            exists (str): Behavior if the target exists: 'append', 'replace', or 'error'.
            on_error (callable, optional): Error handler callback.
        """
        super().__init__(step_name=step_name,
                         dataframes=dataframe,
                         exists=exists,
                         func=self.func,
                         on_error=on_error)
        self.dataframe = dataframe
        self.target = target
        self.conn = engine
        self.schema = schema
        self.kwargs = kwargs

    def func(self, context):
        """
        Execute the load operation. Depending on 'exists', may drop and recreate the table,
        raise on conflict, or append. Inserts all rows of the DataFrame in a single batch.

        Args:
            context: Pipeline context containing dataframes.

        Returns:
            context: Updated Pipeline context after loading.

        Raises:
            ValueError: If exists='error' and target table already exists.
        """
        df = context.get_dataframe(self.dataframe)
        full_name = f"[{self.schema}].[{self.target}]" if self.schema else f"[{self.target}]"

        if not self._table_exists(full_name):
            self._create_table(full_name, df)

        if self.exists == "replace":
            self._drop_table_if_exists(full_name)
            self._create_table(full_name, df)
        elif self.exists == "error" and self._table_exists(full_name):
            raise ValueError(f"Target table {full_name} already exists")

        self._insert_rows(full_name, df)
        return context

    def _table_exists(self, full_name: str) -> bool:
        """
        Check if the target table exists in INFORMATION_SCHEMA.TABLES.

        Args:
            full_name (str): Qualified table name.

        Returns:
            bool: True if table exists, False otherwise.
        """
        schema, table = full_name.strip("[]").split("].[")
        sql = (
            "SELECT 1 FROM INFORMATION_SCHEMA.TABLES "
            "WHERE TABLE_SCHEMA = ? AND TABLE_NAME = ?"
        )
        cursor = self.conn.cursor()
        cursor.execute(sql, schema, table)
        exists = cursor.fetchone() is not None
        cursor.close()
        return exists

    def _drop_table_if_exists(self, full_name: str):
        """
        Drop the target table if it exists.

        Args:
            full_name (str): Qualified table name.
        """
        cursor = self.conn.cursor()
        cursor.execute(f"IF OBJECT_ID('{full_name}', 'U') IS NOT NULL DROP TABLE {full_name}")
        self.conn.commit()
        cursor.close()

    def _create_table(self, full_name: str, df: pd.DataFrame):
        """
        Create a new table with column definitions inferred from DataFrame dtypes.

        Args:
            full_name (str): Qualified table name.
            df (pd.DataFrame): DataFrame whose schema will be used.
        """
        cols = [f"[{col}] {self._map_dtype(dtype, df[col])}" for col, dtype in df.dtypes.items()]
        ddl = f"CREATE TABLE {full_name} ({', '.join(cols)})"
        cursor = self.conn.cursor()
        cursor.execute(ddl)
        self.conn.commit()
        cursor.close()

    def _map_dtype(self, dtype, series) -> str:
        """
        Map a pandas dtype to an appropriate SQL column type.

        Args:
            dtype: pandas dtype of the column.
            series (pd.Series): Column data for length inference.

        Returns:
            str: SQL datatype declaration.
        """
        if pd.api.types.is_integer_dtype(dtype):
            return "BIGINT"
        if pd.api.types.is_float_dtype(dtype):
            return "FLOAT"
        if pd.api.types.is_bool_dtype(dtype):
            return "BIT"
        if pd.api.types.is_datetime64_any_dtype(dtype):
            return "DATETIME2"
        max_len = series.astype(str).str.len().max()
        return f"NVARCHAR({int(max_len)})" if max_len and max_len < 4000 else "NVARCHAR(MAX)"

    def _insert_rows(self, full_name: str, df: pd.DataFrame):
        """
        Insert DataFrame rows into the target table using executemany for performance.

        Args:
            full_name (str): Qualified table name.
            df (pd.DataFrame): DataFrame to insert.
        """
        cols = ", ".join(f"[{c}]" for c in df.columns)
        placeholders = ", ".join("?" for _ in df.columns)
        sql = f"INSERT INTO {full_name} ({cols}) VALUES ({placeholders})"
        data = df.values.tolist()

        cursor = self.conn.cursor()
        cursor.fast_executemany = True
        cursor.executemany(sql, data)
        self.conn.commit()
        cursor.close()

    def map_exists_parameter(self):
        """
        Public method: map_exists_parameter()
        Maps the exists parameter, No mapping required for this instance

        Returns:
            str (['error', 'replace', 'append'], None): 
        """
        return self.exists



"""
Module: distributed_chunker

This module implements the DistributedChunker class, a subclass of the Chunker abstract class.
DistributedChunker calculates chunk coordinates using a distributed strategy.
It computes the total number of chunks by multiplying the number of chunks available for each
step and then evenly distributes rows among those chunks.
This approach ensures a balanced allocation of rows across all chunks for efficient processing.
"""

import math
from .chunker import Chunker

class DistributedChunker(Chunker):
    """
    DistributedChunker

    A concrete implementation of the Chunker class that calculates chunk coordinates using a
    distributed strategy. Instead of processing chunks in a round-robin manner, this chunker
    computes the total number of chunks based on the individual chunk sizes of the steps and
    then calculates the start and end indices for each chunk, ensuring an even distribution.
    """
    def __init__(self, step_index):
        """
        Distributed Chunker Class Constructor
        Initializes the DistributedChunker object by invoking the parent class constructor
        and then invoking the calculate_chunks() method to populate the coordinate queue.

        Arguments:
            step_index (OrderedDict): 
                An ordered dictionary mapping step keys to step objects in the Pipeline.
        """
        super().__init__(step_index=step_index)

    def calculate_chunks(self):
        """
        Public method: calculate_chunks()
        Calculates the chunk coordinates for each step in the chunk index.
        The chunk coordinates are added to the coordinate queue.
        Recursive Chunker calculates the chunk coordinates by first calculating the total
        number of chunks and then iterating through each chunk to calculate the start index
        and number of rows for each chunk.
        By using the total number of chunks, the method calculates the base number of rows
        per chunk and the remainder. This way, the method can distribute the remainder 
        among the chunks, resulting in an even distribution of rows.
        
        This version produces tuples of (start_index, nrows) for
        compatibility with pandas read_csv.
        """
        chunk_keys = list(self.chunk_index.keys())

        chunks_per_key = {}
        for key in chunk_keys:
            chunk_size, _, num_rows, _ = self.chunk_index[key]
            chunks_per_key[key] = math.ceil(num_rows / chunk_size)

        total_chunks = 1
        for key in chunk_keys:
            total_chunks *= chunks_per_key[key]

        for chunk in range(total_chunks):
            for key in chunk_keys:
                _, _, num_rows, _ = self.chunk_index[key]
                base = num_rows // total_chunks
                remainder = num_rows % total_chunks
                start_idx = chunk * base + min(chunk, remainder)
                end_idx = start_idx + base + (1 if chunk < remainder else 0)
                nrows = end_idx - start_idx
                if nrows == 0:
                    break
                self.coordinate_queue.put((start_idx, nrows))



"""
Module: seroflow

This module implements a data pipeline framework using the Pipeline class.
It provides functionality for managing and executing a sequence of data processing steps
with support for caching, logging, parameter management, and execution chunking.

The module integrates with several sub-components such as custom logging, caching,
context management, transformation utilities, chunking, and type validation
to offer a robust and extensible architecture.

Classes:
    Pipeline: Core class to construct, manage, and execute an ETL Pipeline.
              It enables the addition of steps, manages dependencies via a
              global context and parameter index, and supports various execution modes.
"""
import logging
import time
from collections import OrderedDict
from tqdm.auto import tqdm

from .log import CustomLogger
from .cache import AbstractCache, LFUCache
from .context import Context as base_context
from .transform import CacheState, ReloadCacheState, ResetCache
from .chunker import Chunker
from .wrappers import log_error
from .utils import generate_key
from .types import is_step, is_extractor, is_multiextractor
from .types import is_loader, is_context, is_context_object

class Pipeline():
    """
    Pipeline Class

    The Pipeline class provides a framework for constructing and executing data pipelines 
    with built-in support for state caching, logging, and chunked execution.
    It allows dynamic addition of processing steps while managing inter-step dependencies
    through a global context and parameter index. The Pipeline can operate in different modes:
    ("DEV", "PROD") which alter its execution behavior (e.g: DEV mode skips loader steps).

    Key Features:
        - Dynamically add individual or multiple steps to the Pipeline.
        - Manage shared parameters and dataframes across steps via a global context.
        - Support for caching intermediate states to enable execution resumption.
        - Integration with custom or user-defined logging mechanisms to record events.
        - Optional support for chunking, allowing segmented processing of large datasets.
        - Automatic validation and indexing of extractors, loaders, and step outputs.
        - Flexible configuration through properties and setters ensuring type safety and 
          proper initialization.

    Attributes:
        logger (logging.Logger or None): 
            Logger instance used for tracking Pipeline execution details.
        mode (str): 
            The current execution mode that affects how steps are processed.
        cache (AbstractCache or None): 
            Caching mechanism for storing intermediate Pipeline states.
        parameter_index (dict): 
            Dictionary for storing parameters and variables shared across steps.
        step_index (OrderedDict): 
            Ordered dictionary mapping unique step keys to their corresponding step objects.
        step_name_index (OrderedDict): 
            Ordered dictionary mapping unique step keys to the step names.
        dataframe_index (dict): 
            Dictionary mapping step keys to requested dataframe names.
        globalcontext (Context): 
            Global context object that holds all dataframes used throughout the Pipeline.
        chunker (Chunker or None): 
            Optional chunker object to partition and manage segmented execution.
    """

    def __init__(self, cache=False, logger=False, mode="DEV"):
        """
        Pipeline Class Constructor method.

        Arguments: 
            logger (python.logging, Bool): 
                Default: False
                    Does not instantiate a logger, no logging used.
                True:
                    Instantiates default logger: CustomLogger
                python.logging:
                    User defined python.logging object.
            cache (AbstractCache subclass, Bool):
                Default: False
                    Does not instantiate a cache, no cache used.
                True:
                    Instantiates default cache: LFUCache
                AbstractCache subclass:
                    User defined Cache object must be derived from AbstractCache class.
            mode (str: ["DEV", "PROD"]):
                Default: "DEV"
                    Instantiates Pipeline as Development Mode
                "PROD":
                    Instantiates Pipeline as Production Mode

        """
        self.logger = logger
        self.mode = mode # DEV, PROD
        self.checked_targets = False
        self.globalcontext = base_context("globalcontext")
        self.cache = cache
        self.__target_extractor = None
        self.__target_loader = None
        self.__chunker = None
        self.__parameter_index = {}
        self.__step_index = OrderedDict()
        self.__step_name_index = OrderedDict()
        self.__dataframe_index = {}

    def __del__(self):
        """
        Pipeline Object Destructor method.
        Deletes necessary components on Program Completion or Object Deletion.
        """
        del self.__logger
        del self.__target_extractor
        del self.__target_loader
        del self.__chunker
        del self.__parameter_index
        del self.__step_index
        del self.__step_name_index
        del self.__dataframe_index
        del self.__globalcontext
        del self.__cache

    def __str__(self):
        """
        Pipeline Object Custom Print method.
        Used to display internal Parameter Index, Step Index and Dataframe Index.
        """
        print("----Seroflow Pipeline----")
        print(f"Parameters Index: {self.parameter_index}")
        print(f"Step Index: {self.step_index}")
        print(f"Step Name Index: {self.step_name_index}")
        print(f"Dataframe Index: {self.dataframe_index}")
        return "-----------------------"

    def __display_message(self, message, _print=False):
        """
        Custom display method.
        Used to display messages, if necessary or log messages if logger is initialized.

        Arguments:
            message (str): 
                Formatted message to be printed, logged or both.
            _print (bool):
                True:
                    Message is printed.
                False:
                    Message is not printed.
        """
        if self.logger_is_set():
            self.logger.info(message)
        if _print:
            print(message)

    @property
    def logger(self):
        """
        Logger Property Getter Method.
        Logger Property (Optional):
        Used to continuously log any execution info or errors.
        Works with @log_error() wrapper to wrap callable in try-except and log any errors.

        Returns:
            self.__logger
        """
        return self.__logger

    @property
    def target_extractor(self):
        """
        Target Extractor Property Getter Method.
        Target Extractor Property (Required):
        Used to verify that Pipeline execution always begins on an Extract Step.
        Ensures that there is always some data in execution of the Pipeline object.

        Returns:
            self.__target_extractor
        """
        return self.__target_extractor

    @property
    def target_loader(self):
        """
        Target Loader Property Getter Method.
        Target Loader Property (Optional):
        Used to verify that data is loaded as final step.

        Returns:
            self.__target_loader
        """
        return self.__target_loader

    @property
    def cache(self):
        """
        Cache Property Getter Method.
        Cache Property (Optional):
        Stores cache object.
        Used to cache any steps in development or in execution.
        In execution, cache can be used to 'branch' steps by saving the current state.
        In development, cache is used to store the state on a completed step so that
        on rerun, execution continues from last completed step.

        Returns:
            self.__cache
        """
        return self.__cache

    @property
    def parameter_index(self):
        """
        Parameter Index Property Getter Method.
        Parameter Index Property (Required):
        Used to store parameters/variables to be used across steps.
        Step Functions can pass variables to be saved through the return statement.
        Pipeline object will read the return statement,
        then save/update the value of the variable inside parameter index.
        * Steps are not required to have parameters

        Mapping:
            {
                Parameter name: Parameter value,
                ...
            }

        Returns:
            self.__parameter_index
        """
        return self.__parameter_index

    @property
    def step_index(self):
        """
        Step Index Property Getter Method.
        Step Index Property (Required):
        Used to store Step Objects.
        Step Objects are added using the add_step() or add_steps() public methods.
        Step Objects are then verified, and pass through the parse_step() public method.
        Unique Step key is created by hashing: step name (often function name) and step num.
        Step Object and Key is then stored for future use in step index.
        * All Step Objects are required to be added to step index.

        Mapping:
            {
                Step Key: Step Object,
                ...
            }

        Returns:
            self.__step_index
        """
        return self.__step_index

    @property
    def step_name_index(self):
        """
        Step Name Index Property Getter Method.
        Step Name Index Property (Required):
        Used to Map Step Names to Step Keys.
        Unique Step key is created by hashing: step name (often function name) and step num.
        Step Name and Key is then stored for future use in step name index.
        * All Step Objects are required to be added to step name index.

        Mapping:
            {
                Step Key: Step Name,
                ...
            }

        Returns:
            self.__step_name_index
        """
        return self.__step_name_index

    @property
    def dataframe_index(self):
        """
        DataFrame Index Property Getter Method.
        DataFrame Index Property (Required):
        Used to Map Step Keys with their requested Dataframes.
        Step Key and corresponding list of dataFrame names is stored.
        * Steps are not required to list dataframes needed, 
        * however, if 'context' argument is found in step function signature
        then global context is given.

        Mapping:
            {
                Step Key: [DataFrame Name, DataFrame Name, ...],
                ...
            }

        Returns:
            self.__dataframe_index
        """
        return self.__dataframe_index

    @property
    def globalcontext(self):
        """
        Global Context Property Getter Method.
        Global Context Property (Required):
        Stores context object.
        Used to store all dataframes in single context object.
        Pipeline will retrieve requested dataframe from global context if specified by step.
        Or, will return entire global context (ie all dataframes).

        Returns:
            self.__globalcontext
        """
        return self.__globalcontext

    @property
    def chunker(self):
        """
        Chunker Property Getter Method.
        Chunker Property (Optional):
        Stores chunker object.
        Used to store/initialize necessary properties to perform chunking on Pipeline execution.
        Chunker is passed in the pipeline.execute(chunker=chunker) method.

        Returns:
            self.__chunker
        """
        return self.__chunker

    @property
    def mode(self):
        """
        Mode Property Getter Method.
        Mode Property (Optional):
        Used to determine current execution Mode: "DEV" or "PROD"
        Pipeline Execution will change depending on mode.
        DEV:
            Loader Steps are skipped so that data can first be tested without actually loading.
        PROD:
            All Loader Steps are executed.

        Returns:
            self.__mode
        """
        return self.__mode

    @logger.setter
    @log_error("Logger must be a logging object")
    def logger(self, logger):
        """
        Logger Property Setter Method.
        Receives argument logger.
        Verifies that argument logger is type Bool or python.logging.

        Arguments:
            logger (python.logging, Bool):
                True:
                    Default CustomLogger is initialized.
                False:
                    No logger is initialized.
                python.logging:
                    Predefined logger is initialized.

        Raises:
            TypeError:
                Argument logger is not of type Bool or python.logging.
        """
        if not logger:
            self.__logger = None
        elif isinstance(logger, logging.Logger):
            self.__logger = logger
            self.__display_message("Logger set...")
        elif logger is True:
            self.__logger = CustomLogger("Pipeline").logger
            self.__display_message("Logger set...")
        else:
            raise TypeError("Logger must be a logging object")

    @target_extractor.setter
    @log_error("Verify Extractor Type")
    def target_extractor(self, extractor):
        """
        Target Extractor Property Setter Method.
        Receives argument extractor.
        Verifies that argument extractor is type Extractor or MultiExtractor.

        Arguments:
            extractor (Extractor, MultiExtractor):
                Initializes Target Extractor property.

        Raises:
            TypeError:
                Argument extractor is not of type Extractor or MultiExtractor.
        """
        if is_extractor(extractor, _raise=False):
            self.__target_extractor = extractor
        elif is_multiextractor(extractor, _raise=False):
            self.__target_extractor = extractor
        else:
            raise TypeError("Extractor must be extractor or multiextractor")
        self.__display_message("Target extractor set...")

    @target_loader.setter
    @log_error("Verify Loader Type")
    def target_loader(self, loader):
        """
        Target Loader Property Setter Method.
        Receives argument loader.
        Verifies that argument loader is type Loader.

        Arguments:
            loader (Loader):
                Initializes Target Loader property.

        Raises:
            TypeError:
                Argument loader is not of type Loader.
        """
        if is_loader(loader, _raise=True):
            self.__target_loader = loader
            self.__display_message("Target loader set...")

    @parameter_index.setter
    @log_error("Parameter index must be a dictionary")
    def parameter_index(self, parameter_index):
        """
        Parameter Index Property Setter Method.
        Receives argument parameter_index.
        Verifies that parameter_index loader is type Dict.

        Arguments:
            parameter_index (Dict):
                Initializes Parameter Index property.

        Raises:
            TypeError:
                Argument parameter_index is not of type Dict.
        """
        if not isinstance(parameter_index, dict):
            raise TypeError("Parameter index must be a dictionary")
        self.__parameter_index = parameter_index

    @globalcontext.setter
    @log_error("Global context must be a context object")
    def globalcontext(self, globalcontext):
        """
        Global Context Property Setter Method.
        Receives argument globalcontext.
        Verifies that globalcontext is type Context.

        Arguments:
            globalcontext (Context):
                Initializes Global Context property.

        Raises:
            TypeError:
                Argument globalcontext is not of type Context.
        """
        if not is_context(globalcontext):
            raise TypeError("Global context must be a Context object")
        self.__globalcontext = globalcontext

    @cache.setter
    @log_error("Cache must be an instance of AbstractCache")
    def cache(self, cache):
        """
        Cache Property Setter Method.
        Receives argument cache.
        Verifies that argument cache is type Bool or subclass to AbstractCache.

        Arguments:
            cache (subclass to AbstractCache, Bool):
                True:
                    Default LFUCache is initialized.
                False:
                    No cache is initialized.
                subclass to AbstractCache:
                    Predefined cache is initialized.

        Raises:
            TypeError:
                Argument cache is not of type Bool or subclass to AbstractCache.
        """
        if not cache:
            self.__cache = None
            self.__display_message("Cache not set...")
        elif isinstance(cache, AbstractCache):
            self.__cache = cache
            self.__display_message("Cache set...")
        elif cache is True:
            self.__cache = LFUCache()
            self.__display_message("Cache set...")
        else:
            raise TypeError("Cache must be type AbstractCache or Bool")

    @mode.setter
    @log_error("Mode must be either DEV, or PROD")
    def mode(self, mode):
        """
        Mode Property Setter Method.
        Receives argument mode.
        Verifies that mode is type string and value: "DEV" or "PROD".

        Arguments:
            mode (str):
                "DEV":
                    Initializes Mode to value "DEV".
                "PROD":
                    Initializes Mode to value "PROD".

        Raises:
            TypeError:
                Argument mode is not of type str or has value: "DEV" or "PROD".
        """
        if not isinstance(mode, str):
            raise TypeError("Mode must be a string")
        if mode not in ["DEV", "PROD"]:
            raise ValueError("Mode must be either DEV, or PROD")
        self.__mode = mode

    @chunker.setter
    @log_error("Chunker must be of Chunker class type")
    def chunker(self, chunker):
        """
        Chunker Property Setter Method.
        Receives argument chunker.
        Verifies that argument chunker is a class and a subclass of Chunker.
        Adds a reset cache step if a cache is also initialized.
        Adding Reset ensures that cache is not used on chunks.
        Saves current state of global context and parameter index.

        Arguments:
            chunker (subclass of Chunker):
                A class that is a subclass of Chunker.

        Raises:
            TypeError:
                If the provided chunker is not a subclass of Chunker.
        """
        if not (isinstance(chunker, type) and issubclass(chunker, Chunker)):
            raise TypeError("Chunker must be a subclass of Chunker")
        if self.__cache_is_set():
            self.add_step(self.reset_cache(delete_directory=True))
        if self.chunker is None:
            self.__chunker = chunker(self.step_index)
            self.__chunker.save(parameter_index=self.parameter_index,
                                globalcontext=self.globalcontext)
            self.__display_message("Chunker initialized...")

    def logger_is_set(self):
        """
        Public Method: logger_is_set()
        Verifies if logger is initialized.
        Must be public as wrapper, @log_error() uses method.

        Returns (Bool):
            True:
                Logger is initialized.
            False:
                Logger is not initialized.
        """
        if not self.logger:
            return False
        return True

    def __cache_is_set(self):
        """
        Private Method: __cache_is_set()
        Verifies if cache is initialized.

        Returns (Bool):
            True:
                Cache is initialized.
            False:
                Cache is not initialized.
        """
        if not self.cache:
            return False
        return True

    def __chunker_is_set(self):
        """
        Private Method: __chunker_is_set()
        Verifies if chunker is initialized.

        Returns (Bool):
            True:
                Chunker is initialized.
            False:
                Chunker is not initialized.
        """
        if not self.chunker:
            return False
        return True

    def __update_parameter_index(self, parameter, value):
        """
        Private Method: __update_parameter_index()
        Updates parameter index values.

        Arguments:
            parameter (string):
                Parameter Name to be updated.
            value (Any):
                Parameter value to add/update.
        """
        self.parameter_index[parameter] = value

    def __add_new_parameter(self, parameter):
        """
        Private Method: __add_new_parameter()
        Adds new empty parameter to parameter index.
        Used in __parse_step() method.

        Arguments:
            parameter (string):
                Parameter Name to be added.
        """
        if parameter not in self.parameter_index:
            self.parameter_index[parameter] = None

    def __update_step_index(self, step_key, step):
        """
        Private Method: __update_step_index()
        Updates step index values.

        Arguments:
            step_key (string):
                Step Key to be updated.
            step (Step Object):
                Step Object to add/update.
        """
        self.step_index[step_key] = step

    def __update_step_name_index(self, step_key, step_name):
        """
        Private Method: __update_step_name_index()
        Updates step name index values.

        Arguments:
            step_key (string):
                Step Key to be updated.
            step_name (string):
                Step Name to add/update.
        """
        self.step_name_index[step_key] = step_name

    def __update_dataframe_index(self, step_key, dataframe_name):
        """
        Private Method: __update_dataframe_index()
        Updates dataframe index values.

        Arguments:
            step_key (string):
                Step Key to be updated.
            dataframe_name (string):
                Dataframe name to add/update.
        """
        if step_key not in self.dataframe_index:
            self.dataframe_index[step_key] = []
        self.dataframe_index[step_key].append(dataframe_name)

    def __update_globalcontext(self, subcontext):
        """
        Private Method: __update_globalcontext()
        Updates global context dataframes.
        Either Updates internal dataframe.
        Or, Adds new dataframe to global context.

        Arguments:
            subcontext (Context):
                Subcontext, contains dataframes to update in global context.
        """
        for dataframe_name in subcontext.get_dataframe_names():
            if dataframe_name in self.globalcontext.get_dataframe_names():
                self.globalcontext.set_dataframe(
                    dataframe_name, subcontext.get_dataframe(dataframe_name)
                )
            else:
                self.globalcontext.add_dataframe(
                    dataframe_name, subcontext.get_dataframe(dataframe_name)
                )

    def __update_cache(self, step_key):
        """
        Private Method: __update_cache()
        Updates cache values.
        Once step is completed, cache is then used to store current state of Pipeline.

        Arguments:
            step_key (string):
                Parameter Name to be updated.
        """
        if self.__cache_is_set() and (not isinstance(self.step_index[step_key], ResetCache)):
            self.__store_in_cache(step_key)

    def __check_parsed_parameters(self, kwargs):
        """
        Private Method: __check_parsed_parameters()
        Validates created kwargs to be passed in step function.
        Ensures parameters do not contain empty values.

        Arguments:
            kwargs (dict):
                Mapped parameters and values to be passed into step function.
        """
        for key, value in kwargs.items():
            if value is None:
                error_message = f"Key: {key} has no value, check parameter index"
                raise ValueError(error_message)

    def __check_step_output(self, step_output, step_key):
        """
        Private Method: __check_step_output()
        Validates step output with intended step return list.
        Ensures number of returned values matches expected number of returned values.
        Returns step output as list or None.

        Arguments:
            step_output (tuple):
                Output generated by step function.
            step_key (string):
                Step Key for corresponding step function that was just executed.

        Returns:
            step_output (list):
                Verified step output as a list.
            None:
                if no return statement then nothing should be returned.
        Raises:
            ValueError:
                Number of Elements returned by step function 
                does not match expected number of return elements.
        """
        if not isinstance(step_output, tuple):
            step_output = [step_output]

        if (self.step_index[step_key].return_list == []) and (step_output[0] is None):
            return None
        if len(self.step_index[step_key].return_list) != len(step_output):
            raise ValueError("Error incorrect amount of return elements found")
        return step_output

    def __create_subcontext(self, step, step_key):
        """
        Private Method: __create_subcontext()
        Creates subcontext to be passed to step function.
        Subcontext contains only the dataframes needed from the global context.
        If step is of extractor type, then blank subcontext is returned.

        Arguments:
            step (Step Object):
                Step Object requesting dataframes.
            step_key (string):
                Step Key for corresponding step.

        Returns:
            subcontext (Context):
                Context object containing only the requested dataframes.
        """
        step_name = self.step_name_index[step_key]
        subcontext = base_context(step_name + "_subcontext")
        if not is_extractor(step, _raise=False):
            desired_dataframes = self.dataframe_index.get(step_key, [])
            if not desired_dataframes:
                subcontext = self.globalcontext
            else:
                for dataframe_name in desired_dataframes:
                    subcontext.add_dataframe(
                        dataframe_name,
                        self.globalcontext.get_dataframe(dataframe_name)
                    )
        return subcontext

    def __get_current_step_number(self):
        """
        Private Method: __get_current_step_number()
        Calculates the step number for newest step.

        Returns:
            (Int):
                Number of Steps + 1, or 0 if No steps have been added.
        """
        return self.__get_number_of_steps() + 1 if self.step_index else 0

    def __get_number_of_steps(self):
        """
        Private Method: __get_number_of_steps()
        Calculates the number of steps.

        Returns:
            (Int):
                Number of Steps.
        """
        return len(self.__get_step_keys())

    def __get_step_keys(self):
        """
        Private Method: __get_step_keys()
        Gathers all step keys.

        Returns:
            (List):
                List of step keys.
        """
        return list(self.step_index.keys())

    @log_error("Error adding target to step index")
    def __add_target_to_step(self, target, last=False):
        """
        Private Method: __add_target_to_step()
        Adds targets to step index.

        Arguments:
            target (Extractor, Loader):
                Target Extractor/Loader to be added to step index.
            last (Bool):
                True:
                    target is added to end of step index.
                    Intended to be used with target loader.
                False:
                    target is added to beginning of step index.
                    Intended to be used with target extractor.
        """
        target_key = self.__parse_step(target)
        self.step_index.move_to_end(target_key, last=last)
        self.step_name_index.move_to_end(target_key, last=last)
        self.__display_message(f"Successfully added step with key: {target_key}")

    @log_error("Error adding targets to step index")
    def __add_targets(self):
        """
        Private Method: __add_targets()
        Verifies Targets before adding them to step index.
        If targets have already been checked then method returns.
        In "DEV" Mode target loader and extractor do not have to be set.
        Ensures that MultiExtractors are parsed, by adding single extractor steps.

        Raises:
            ValueError:
                If in "PROD" mode then extractor must be set.
        """
        if not self.checked_targets:
            self.checked_targets = True
            if (self.mode != "DEV") and (not self.target_extractor):
                raise ValueError("Target extractor must be set before executing")
            if self.target_extractor:
                if is_multiextractor(self.target_extractor):
                    for extractor in self.target_extractor.extractors:
                        self.__add_target_to_step(extractor, last=False)
                else:
                    self.__add_target_to_step(self.target_extractor, last=False)
            if self.target_loader:
                self.__add_target_to_step(self.target_loader, last=True)
            self.__display_message("Successfully added targets to steps")

    @log_error("Error Parsing Step")
    def __parse_step(self, step):
        """
        Private Method: __parse_step()
        Creates step key and adds new step to step index and step name index.
        Updates parameter and dataframe index with new values.

        Arguments:
            step (Step Object):
                New step object to be added to indexes.

        Returns:
            step_key (string):
                Newly generated unique key for added step.
        """
        key_index = self.__get_current_step_number()
        step_key = generate_key(f"{step.step_name}_{key_index}")
        self.__update_step_index(step_key, step)
        self.__update_step_name_index(step_key, step.step_name)
        for param in step.params_list:
            self.__add_new_parameter(param)
        for dataframe in step.dataframes:
            self.__update_dataframe_index(step_key, dataframe)
        return step_key

    @log_error("Error Parsing Parameters, proper parameter value not found")
    def __parse_parameters(self, step_key):
        """
        Private Method: __parse_parameters()
        Retrieves needed parameters and subcontext for step function.
        If chunker is used, then chunking coordinates (ie start position and nrows)
        for current chunk is retrieved.
        Gathers parameter value using hierarchy:
            1. Inputted value set in step object instantiation.
            2. Current value inside parameter index.
            3. Default value set in step object instantiation.

        Arguments:
            step_key (string):
                Step Key for current step function about to be executed.

        Returns:
            kwargs (dict):
                Mapped parameter values and subcontext requested by step function.
        """
        kwargs = {}
        step = self.step_index[step_key]

        if step.needs_context:
            subcontext = self.__create_subcontext(step, step_key)
            kwargs["context"] = subcontext

        if ((self.__chunker_is_set()) and
            (is_extractor(step, _raise=False)) and
            (hasattr(step, "chunk_size"))):
            skiprows, nrows = self.chunker.dequeue()
            step.kwargs['skiprows'] = skiprows
            step.kwargs['nrows'] = nrows

        for param in step.params_list:
            input_value = step.input_params.get(param)
            curr_value = self.parameter_index.get(param)
            default_value = step.default_params.get(param)
            param_value = input_value or curr_value or default_value
            kwargs[param] = param_value
        self.__check_parsed_parameters(kwargs)
        return kwargs

    @log_error("Error Parsing Step Output")
    def __parse_step_output(self, step_output, step_key):
        """
        Private Method: __parse_step_output()
        Receives step outputs, validates step output, then updates indexes.
        Updates global context and parameter index.

        Arguments:
            step_output (tuple):
                Output generated by step function.
            step_key (string):
                Step Key for corresponding step function that was just executed.
        """
        checked_output = self.__check_step_output(step_output, step_key)
        if checked_output is None:
            return
        for param, value in zip(self.step_index[step_key].return_list, checked_output):
            if is_context(value):
                self.__update_globalcontext(value)
            elif is_context_object(value):
                for _, item in value.items():
                    self.__update_globalcontext(item)
            else:
                self.__update_parameter_index(param, value)

    def __load_from_cache(self, step_keys):
        """
        Private Method: __load_from_cache()
        On Pipeline execution, if a cache is set then,
        the state of the Pipeline object at the last cached completed step is loaded.
        The starting step's index is returned.
        If cache is not set or there is no cached checkpoint, execution starts at step 0.

        Arguments:
            step_keys (list):
                List of step keys currently in step index.

        Returns:
            start_index (int):
                position of first step to be executed.
        """
        start_index = 0
        cached_checkpoint = self.cache.get_cached_checkpoint(self.step_index)
        if cached_checkpoint and (cached_checkpoint in step_keys):
            self.parameter_index, self.globalcontext = self.cache.load(cached_checkpoint)
            start_index = step_keys.index(cached_checkpoint) + 1
            self.__display_message(f"Resuming execution from: {cached_checkpoint}", True)
        else:
            self.__display_message("No checkpoint found, starting from beginning...", True)
        return start_index

    def __store_in_cache(self, step_key):
        """
        Private Method: __store_in_cache()
        If cache is set, then once a step has fully completed,
        the state of execution at that point is cached.

        Arguments:
            step_key (string):
                Step Key for completed step
        """
        self.cache.store(self.step_index, self.parameter_index, self.globalcontext, step_key)
        self.__display_message(f"Checkpoint stored for step: {step_key}")

    @log_error("add_steps method requires a list of step objects")
    def add_steps(self, steps):
        """
        Public Method: add_steps()
        Adds list of steps to Pipeline indexes.

        Arguments:
            steps (list):
                list of step objects to be addded.

        Raises:
            TypeError:
                steps argument must be of type list.
        """
        if not isinstance(steps, list):
            raise TypeError("try using a list...")
        for step in steps:
            self.add_step(step)

    @log_error("add_step method requires a step object")
    def add_step(self, step):
        """
        Public Method: add_step()
        Adds single step to Pipeline indexes.
        If a multiextractor step is added then verifies internal steps.

        Arguments:
            step (Step Object):
                step object to be addded.

        Raises:
            TypeError:
                step argument must be of type Step.
        """
        if is_multiextractor(step):
            for extractor in step.extractors:
                self.add_step(extractor)
        if is_step(step, _raise=True):
            step_key = self.__parse_step(step)
            self.__display_message(f"Successfully added: {step.step_name}, key: {step_key}")

    def cache_state(self, step_name="cache_state"):
        """
        Public Method: cache_state()
        Intended to be used as a step itself and added using public methods:
        add_step() or add_step().
        Caches the current state of Pipeline object in execution.
        Can be used to 'branch' steps.

        Arguments:
            step_name (string):
                Default: "cache_state"
                Name for added cache state step.

        Returns:
            CacheState predefined step, compatible with any cache object 
            derived from AbstractCache class.
        """
        return CacheState(
            step_name=step_name,
            cache=self.cache,
            parameter_index=self.parameter_index,
            globalcontext=self.globalcontext,
        )

    def reload_cached_state(self, cache_key, step_name="reload_cached_state"):
        """
        Public Method: reload_cached_state()
        Intended to be used as a step itself and added using public method:
        add_step() or add_step().
        Reloads the state of cached state in Pipeline execution.
        Can be used to 'branch' steps.

        Arguments:
            step_name (string):
                Default: "reload_cached_state"
                Name for added reload cache state step.
            cache_key (int):
                index for cached state, configuration depends on cache class used.

        Returns:
            ReloadCacheState predefined step, compatible with any cache object 
            derived from AbstractCache class.
        """
        return ReloadCacheState(
            step_name=step_name,
            cache_key=cache_key,
            cache=self.cache,
            pipeline=self
        )

    def reset_cache(self, step_name="reset_cache", delete_directory=False):
        """
        Public Method: reset_cache()
        Intended to be used as a step itself and added using public method:
        add_step() or add_step().
        Resets the cache in execution.

        Arguments:
            step_name (String):
                Default: "reset_cache"
                Name for added reset cache step.
            delete_directory (Bool):
                True:
                    cache directory is deleted.
                False:
                    cache directory is not deleted.

        Returns:
            ResetCache predefined step, compatible with any cache object 
            derived from AbstractCache class.
        """
        return ResetCache(
            step_name=step_name,
            cache=self.cache,
            delete_directory=delete_directory
        )

    def __perform_step(self, step_key):
        """
        Private Method: __perform_step()
        Executes the current step, by retrieving necessary parameters 
        and updating indexes with returned values.
        Skips loader steps if Pipeline mode is set to "DEV"
        Updates Cache if set.

        Arguments:
            step_key (string):
                Step key for current step
        """
        if is_loader(self.step_index[step_key], _raise=False) and self.mode == "DEV":
            return
        step_params = self.__parse_parameters(step_key)
        step_output = self.step_index[step_key](**step_params)
        self.__parse_step_output(step_output, step_key)
        self.__update_cache(step_key)

    @log_error("Error executing Pipeline...")
    def execute(self, chunker=None):
        """
        Public Method: execute()
        Initializes Pipeline Execution, performing validation on:
        targets, chunker, cache, logger.
        Displays execution messages, times total execution of all steps.

        Arguments:
            chunker (subclass to Chunker):
                Chunker objects which partitions execution depending on Chunker Type.
        """
        self.__add_targets()
        if chunker is not None:
            self.chunker = chunker

        step_keys = self.__get_step_keys()
        start_index = 0 if not self.__cache_is_set() else self.__load_from_cache(step_keys)

        start_time = time.time()
        curr_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(start_time))
        self.__display_message(f"Beginning ETL Execution at time: {curr_time} ...", True)

        total_steps = self.__get_number_of_steps() - start_index
        with tqdm(total=total_steps, desc="Executing Pipeline") as pbar:
            for step_key in step_keys[start_index:]:
                self.__display_message(f"Executing Step: {self.step_name_index[step_key]} ", True)
                self.__perform_step(step_key)
                self.__display_message(f"Step: {self.step_name_index[step_key]} completed...")
                pbar.update(1)

        if self.chunker is not None:
            if self.chunker.keep_executing:
                self.parameter_index, self.globalcontext = self.chunker.reload()
                self.execute(chunker=chunker)
        else:
            end_time = time.time()
            curr_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(end_time))
            self.__display_message(f"ETL Execution Finished at time: {curr_time} ...", True)
            elapsed_time = end_time - start_time
            if elapsed_time > 1.0:
                self.__display_message(f"Total Execution Time: {elapsed_time} seconds", True)



"""
Module: abstract_step

This module defines the AbstractStep class for creating steps.
AbstractStep provides a common interface for initializing, 
cleaning up, and executing Pipeline steps. Subclasses must implement 
the abstract methods start_step(), stop_step(), and execute() 
to provide step-specific functionality.
"""

from abc import ABC, abstractmethod

class AbstractStep(ABC):
    """
    AbstractStep

    An abstract class for defining the structure of a step. This class extends the ABC class
    and provides a common interface for starting, stopping, and executing the main 
    functionality of a step. Subclasses must implement the abstract methods start_step(),
    stop_step(), and execute() to provide step-specific functionality.
    """

    @abstractmethod
    def start_step(self):
        """
        Abstract method: start_step()
        This method should contain logic to initialize the step before execution.
        """

    @abstractmethod
    def stop_step(self):
        """
        Abstract method: stop_step()
        This method should contain logic to clean up the step after execution has completed.
        """

    @abstractmethod
    def execute(self):
        """
        Abstract method: execute()
        This method should contain the main functionality of the step.
        """



"""
Module that defines the Context dataclass for managing pandas DataFrames and metadata.

The Context class provides a structured way to store and retrieve multiple DataFrames,
manage related metadata, and track DataFrame addresses.
"""

from dataclasses import dataclass, field
from typing import Dict, Any
import pandas as pd


@dataclass
class Context:
    """
    A class to represent a context containing pandas DataFrames and metadata.

    Attributes:
        context_name (str): The name of the context.
        dataframes (Dict[str, pd.DataFrame]): A dictionary mapping names to DataFrame objects.
        metadata (Dict[str, Any]): A dictionary to store metadata about the context.
        dataframe_addr (Dict[str, id]): Dictionary mapping DataFrame names to their ids.
    """
    context_name: str
    dataframes: Dict[str, pd.DataFrame] = field(default_factory=dict)
    metadata: Dict[str, Any] = field(default_factory=dict)
    dataframe_addr: Dict[str, id] = field(default_factory=dict)

    def __post_init__(self):
        """
        Post-initialization processing for the Context instance.

        This method initializes the 'num_dfs' metadata to zero.
        Additional normalization of the context name can be added here.
        """
        self.metadata['num_dfs'] = 0
        # Normalize context name if necessary

    def set_context_name(self, name):
        """
        Set the name of the context.

        Args:
            name (str): The new context name.
        """
        self.context_name = name

    def get_dataframe(self, name):
        """
        Retrieve a DataFrame by its name.

        Args:
            name (str): The name of the DataFrame to retrieve.

        Returns:
            pd.DataFrame or None: 
                The DataFrame associated with the given name, or None if not found.
        """
        return self.dataframes.get(name)

    def set_dataframe(self, name, df):
        """
        Set a DataFrame in the context with the given name.

        Args:
            name (str): The name to assign to the DataFrame.
            df (pd.DataFrame): The DataFrame to store.
        """
        self.dataframes[name] = df

    def get_dataframe_names(self):
        """
        Get the names of all stored DataFrames.

        Returns:
            KeysView[str]: A view of the names of the DataFrames stored in the context.
        """
        return self.dataframes.keys()

    def get_metadata(self, key):
        """
        Retrieve a metadata value by its key.

        Args:
            key (str): The key for the metadata item.

        Returns:
            Any or None: The metadata value associated with the key, or None if not found.
        """
        return self.metadata.get(key)

    def set_metadata(self, key, value):
        """
        Set a metadata value for a given key.

        Args:
            key (str): The key for the metadata item.
            value (Any): The value to be stored.
        """
        self.metadata[key] = value

    def added_dataframe_update_metadata(self):
        """
        Update the metadata to reflect the current number of DataFrames.

        This method updates the 'num_dfs' key in the metadata dictionary to match
        the number of DataFrames currently stored.
        """
        self.metadata['num_dfs'] = len(list(self.dataframes.keys()))

    def add_dataframe(self, name, df):
        """
        Add a new DataFrame to the context and update metadata accordingly.

        This method sets the DataFrame with the given name and then updates the
        metadata to reflect the new total count of DataFrames.

        Args:
            name (str): The name to assign to the DataFrame.
            df (pd.DataFrame): The DataFrame to add.
        """
        self.set_dataframe(name, df)
        self.added_dataframe_update_metadata()

    def delete_dataframe(self, name):
        """
        Delete a DataFrame from the context by its name and update metadata accordingly.

        Args:
            name (str): The name of the DataFrame to delete.
        
        Returns:
            bool: True if the DataFrame was successfully deleted, False otherwise.
        """
        if name in self.dataframes:
            del self.dataframes[name]

            if name in self.dataframe_addr:
                del self.dataframe_addr[name]

            self.added_dataframe_update_metadata()
            return True
        return False

    def __str__(self):
        """
        Return a string representation of the Context.

        This implementation prints the context name, the dataframes, and the metadata.
        Note that it returns an empty string after printing.

        Returns:
            str: An empty string.
        """
        print(self.context_name)
        print(self.dataframes)
        print(self.metadata)
        return ""



"""
Module: excel_loader

This module provides concrete implementations for writing DataFrames to Excel files.
It defines two loader classes:
    - ExcelLoader: A loader that writes a single DataFrame to an Excel file.
    - MultiExcelLoader: A loader that handles writing multiple DataFrames to Excel files.
These classes extend the FileLoader class and implement Excel-specific functionality,
including determining the target file path, handling file modes based on the 'exists'
parameter, and managing Excel writing engines.
"""
import os
import pandas as pd
from ..load.file_loader import FileLoader

class ExcelLoader(FileLoader):
    """
    ExcelLoader

    A concrete loader for writing a DataFrame to an Excel file.
    This class extends FileLoader and provides methods to convert DataFrames into Excel format,
    determine the target file path, and handle file modes to the specified existence parameter.
    """

    def __init__(self,
                 target,
                 dataframe,
                 file_extension=".xlsx",
                 exists="append",
                 step_name="ExcelLoader",
                 on_error=None,
                 **kwargs):
        """
        ExcelLoader Class Constructor
        Initializes the ExcelLoader object.

        Arguments:
            target (str): 
                The target directory where the Excel file will be written
            dataframe (DataFrame): 
                The DataFrame to be written to the Excel file
            file_extension (str): 
                The file extension to use when writing the Excel file
            exists (str): 
                The file mode to use when writing the Excel file
            step_name (str): 
                The name of the step
            on_error (str): 
                The error handling strategy
            **kwargs: 
                Additional keyword arguments for the to_excel() method
        """
        super().__init__(target=target,
                         dataframe=dataframe,
                         exists=exists,
                         func=self.func,
                         step_name=step_name,
                         file_extension=file_extension,
                         on_error=on_error,
                         **kwargs)

    def func(self, context):
        """
        Public method: func()
        Reads the DataFrame from the context and writes it to an Excel file

        Arguments:
            context (Context): 
                The context object containing the dataframes to be written to the Excel
        """
        for key, df in context.dataframes.items():
            if self.target_file_path is None:
                file_path = key+self.file_extension
                target_file_path = os.path.join(self.target, file_path)
            else:
                target_file_path = self.target_file_path
            self.__to_excel(df, target_file_path, self.kwargs)

    def __to_excel(self, df, target_file_path, kwargs):
        """
        Private method: __to_excel()
        Writes the DataFrame to an Excel file

        Arguments:
            df (DataFrame): 
                The DataFrame to be written to the Excel file
            target_file_path (str): 
                The target file path where the Excel file will be written
            kwargs (dict): 
                Additional keyword arguments for the to_excel() method
        """
        directory = os.path.dirname(target_file_path)
        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

        engine = 'openpyxl'
        if self.file_extension == '.xls':
            engine = 'xlrd'

        if os.path.exists(target_file_path):
            with pd.ExcelWriter(target_file_path,
                                engine=engine,
                                mode='a',
                                if_sheet_exists=self.map_exists_parameter()
                                ) as f:
                df.to_excel(f, **kwargs)
        else:
            with pd.ExcelWriter(target_file_path,
                                engine=engine,
                                mode='w'
                                ) as f:
                df.to_excel(f, **kwargs)

    def map_exists_parameter(self):
        """
        Public method: map_exists_parameter()
        Maps the exists parameter to the if_sheet_exists parameter for the ExcelWriter object

        Returns:
            str (['error', 'replace', 'append'], None): 
                The appropriate if_sheet_exists parameter for the ExcelWriter object
        """
        if self.exists == "append":
            return 'overlay'
        if self.exists == "fail":
            return 'error'
        if self.exists == "replace":
            return 'replace'
        return None

class MultiExcelLoader(ExcelLoader):
    """
    MultiExcelLoader

    A concrete loader for writing multiple DataFrames to multiple Excel files.
    This class extends ExcelLoader and provides methods to write multiple DataFrames
    to Excel files in the specified target directory.
    """
    def __init__(self,
                 target,
                 dataframes=None,
                 file_extension=".xlsx",
                 exists="append",
                 step_name="MultiExcelLoader",
                 on_error=None,
                 **kwargs):
        """
        MultiExcelLoader Class Constructor
        Initializes the MultiExcelLoader object.

        Arguments:
            target (str): 
                The target directory where the Excel files will be written
            dataframes (dict): 
                A dictionary mapping keys to DataFrames to be written to the Excel files
            file_extension (str): 
                The file extension to use when writing the Excel files
            exists (str): 
                The file mode to use when writing the Excel files
            step_name (str): 
                The name of the step
            on_error (str): 
                The error handling strategy
            **kwargs: 
                Additional keyword arguments for the to_excel() method
        """
        super().__init__(target=target,
                         dataframe=dataframes,
                         file_extension=file_extension,
                         exists=exists,
                         step_name=step_name,
                         on_error=on_error,
                         **kwargs)



"""
Module: wrappers

This module provides decorators which aid with enhanced logging capabilities.
It includes a timer decorator to measure and log the execution time of functions,
and a log_error decorator that catches exceptions during function execution,
logs error details (such as the function name and line number), and optionally
re-raises a new exception with a custom error message.
"""

from functools import wraps
import time
import traceback

def timer(func):
    """
    Decorator that logs the execution time of the decorated function.
    Execution time is computed by measuring time before and after the function call.
    * only compatible with class methods that have a logger attribute.
    
    Arguments:
        func (function): The function whose execution time will be measured.

    Returns:
        function: The wrapped function with execution time logging.
    """
    @wraps(func)
    def wrap(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        args[0].logger.info("%s took: %s sec", func.__name__, end - start)
        return result
    return wrap

def log_error(err_msg, log_only=False):
    """
    Decorator factory that returns a decorator to log errors during function execution.

    This decorator wraps the target function in a try-except block.
    If an exception occurs, it extracts the location of the error and logs the error
    using the object's logger if it is set. 
    Depending on the 'log_only' flag, it either re-raises a new exception with a custom 
    error message or simply logs the error.

    Arguments:
        err_msg (string):
            Custom error message to be used when re-raising the exception.
        log_only (bool, optional):
            False (Default): re-raises the exception with a custom error message.
            True: only logs the error without re-raising the exception.

    Returns:
        function:
            Decorator that can be applied to a function to add error logging.
    """
    def log_error_inner(func):
        """
        Inner decorator that wraps target function with error logging.

        Arguments:
            func (function):
                The function to be wrapped.

        Returns:
            function:
                The wrapped function that logs errors if they occur.
        """
        @wraps(func)
        def wrap(*args, **kwargs):
            """
            Wraps the decorated function in a try-except block to catch and log exceptions.

            If an exception occurs, logs the function name and the code line of the error.
            If the logger is set on the first argument, logs the error details.
            Depending on the log_only flag, the function either re-raises a new 
            exception with a custom error message or suppresses the exception after logging.

            Returns:
                The result of the function:
                    If no exception occurs
                None:
                    If the exception is logged and suppressed.
            """
            try:
                return func(*args, **kwargs)
            except Exception as e:
                tb_last_frame = traceback.extract_tb(e.__traceback__)[-1]
                _, _, function_name, code_line = tb_last_frame
                if args[0].logger_is_set():
                    raised_msg = f"Error Occurred at: {function_name}; On line: {code_line};"
                    args[0].logger.error(raised_msg)
                    if not log_only:
                        args[0].logger.error(f"Exception {e}")
                        raise RuntimeError(err_msg) from e
                else:
                    raise RuntimeError(err_msg) from e
                return None
        return wrap
    return log_error_inner



"""
Module: column.py

This module implements a collection of transformation classes that perform various column
operations on DataFrames.These transformations include converting column data types,
renaming columns, dropping columns, adding new columns (based on computations or
constant values), merging and splitting columns, and exploding columns. Each transformation
class extends the base Transformation class and is designed to update the DataFrame in
the Pipeline context.

Classes:
    ConvertColumnType: Converts a specified column of a DataFrame to a new data type.
    RenameColumns: Renames one or more columns in a DataFrame based on a provided mapping.
    DropColumn: Drops a single specified column from a DataFrame.
    DropColumns: Drops multiple specified columns from a DataFrame.
    AddColumn: Adds a new column to a DataFrame computed from a function.
    MergeColumns: Merges multiple columns into a single column by concatenating
                  their string representations.
    SplitColumn: Splits a single column into multiple columns based on a delimiter.
    ExplodeColumn: Explodes a column containing lists into multiple rows.
    CreateColumnFromVariable: Creates a new column in a DataFrame using a constant
                              value provided via a variable.
"""

import pandas as pd
from .transformation import Transformation

class ConvertColumnType(Transformation):
    """
    ConvertColumnType

    Converts a specified column of a DataFrame to a new data type.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        column (str): The column to be converted.
        new_type (type): The target data type to convert the column to.
    """
    def __init__(self,
                 dataframe,
                 column,
                 new_type,
                 step_name="ConvertColumnType",
                 on_error=None):
        """
        Initializes the ConvertColumnType transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update.
            column (str): The name of the column whose type is to be converted.
            new_type (type): The target data type for the column.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "ConvertColumnType".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.column = column
        self.new_type = new_type
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the conversion of the column type.

        Retrieves the DataFrame from the context, converts the specified column to the new type,
        updates the DataFrame in the context, and returns the modified context.

        Arguments:
            context (Context): The context containing the DataFrame.

        Returns:
            Context: The updated context with the converted column.
        """
        df = context.dataframes[self.dataframe]
        df[self.column] = self.__convert_column_type(df, self.column, self.new_type)
        context.set_dataframe(self.dataframe, df)
        return context

    def __convert_column_type(self, df, column, new_type):
        """
        Converts the specified column to the new data type.

        Arguments:
            df (DataFrame): The DataFrame containing the column.
            column (str): The column to convert.
            new_type (type): The target data type.

        Returns:
            Series: The converted column.
        """
        return df[column].astype(new_type)


class RenameColumns(Transformation):
    """
    RenameColumns

    Renames one or more columns in a DataFrame based on a provided mapping.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        columns_mapping (dict): A dictionary mapping current column names to new column names.
    """
    def __init__(self,
                 dataframe,
                 columns_mapping,
                 step_name="RenameColumns",
                 on_error=None):
        """
        Initializes the RenameColumns transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update.
            columns_mapping (dict): Mapping of existing column names to new names.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "RenameColumns".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.columns_mapping = columns_mapping
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the column renaming transformation.

        Retrieves the DataFrame from the context, renames the columns based on the mapping,
        updates the DataFrame in the context, and returns the modified context.

        Arguments:
            context (Context): The context containing the DataFrame.

        Returns:
            Context: The updated context with renamed columns.
        """
        df = context.dataframes[self.dataframe]
        context.set_dataframe(self.dataframe, self.__rename_columns(df))
        return context

    def __rename_columns(self, df):
        """
        Renames columns in the DataFrame based on the mapping.

        Arguments:
            df (DataFrame): The DataFrame whose columns are to be renamed.

        Returns:
            DataFrame: The DataFrame with renamed columns.
        """
        # inplace renaming returns None, so we perform renaming and then return the DataFrame.
        df.rename(columns=self.columns_mapping, inplace=True)
        return df


class DropColumn(Transformation):
    """
    DropColumn

    Drops a specified column from a DataFrame.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        column (str): The name of the column to drop.
    """
    def __init__(self,
                 dataframe,
                 column,
                 step_name="DropColumn",
                 on_error=None):
        """
        Initializes the DropColumn transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update.
            column (str): The column to drop.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "DropColumn".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.column = column
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the DropColumn transformation.

        Retrieves the DataFrame from the context, drops the specified column,
        updates the DataFrame in the context, and returns the modified context.

        Arguments:
            context (Context): The context containing the DataFrame.

        Returns:
            Context: The updated context with the column dropped.
        """
        df = context.dataframes[self.dataframe]
        context.set_dataframe(self.dataframe, self.__drop_column(df))
        return context

    def __drop_column(self, df):
        """
        Drops the specified column from the DataFrame.

        Arguments:
            df (DataFrame): The DataFrame to modify.
        
        Returns:
            DataFrame: The DataFrame after dropping the column.
        """
        return df.drop(columns=[self.column])


class DropColumns(Transformation):
    """
    DropColumns

    Drops multiple specified columns from a DataFrame.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        columns (list): A list of columns to drop.
    """
    def __init__(self,
                 dataframe,
                 columns,
                 step_name="DropColumns",
                 on_error=None):
        """
        Initializes the DropColumns transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update.
            columns (list): A list of column names to drop.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "DropColumns".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.columns = columns
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the DropColumns transformation.

        Retrieves the DataFrame from the context, drops the specified columns
        (ignoring errors if columns are missing), updates the DataFrame in the context,
        and returns the modified context.

        Arguments:
            context (Context): The context containing the DataFrame.

        Returns:
            Context: The updated context with the specified columns dropped.
        """
        df = context.dataframes[self.dataframe]
        context.set_dataframe(self.dataframe, self.__drop_columns(df))
        return context

    def __drop_columns(self, df):
        """
        Drops the specified columns from the DataFrame, ignoring errors if column does not exist.

        Arguments:
            df (DataFrame): The DataFrame to modify.
        
        Returns:
            DataFrame: The DataFrame after dropping the columns.
        """
        return df.drop(columns=self.columns, errors='ignore')


class AddColumn(Transformation):
    """
    AddColumn

    Adds a new column to a DataFrame computed from a provided function.
    
    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        column (str): The name of the new column to add.
        compute_func (function): A function that computes the new column's values from DataFrame.
    """
    def __init__(self,
                 dataframe,
                 column,
                 compute_func,
                 step_name="AddColumn",
                 on_error=None):
        """
        Initializes the AddColumn transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update.
            column (str): The name of the new column to add.
            compute_func (function): A function that accepts a DataFrame and returns a
                                     Series representing the new column.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "AddColumn".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.column = column
        self.compute_func = compute_func
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the AddColumn transformation.

        Retrieves the DataFrame from the context, computes the new column values using
        compute_func, adds the new column to the DataFrame, updates the context, and returns it.

        Arguments:
            context (Context): The context containing the DataFrame.

        Returns:
            Context: The updated context with the new column added.
        """
        df = context.dataframes[self.dataframe]
        df[self.column] = self.__compute_column(df)
        context.set_dataframe(self.dataframe, df)
        return context

    def __compute_column(self, df):
        """
        Computes the new column using the provided compute function.

        Arguments:
            df (DataFrame): The DataFrame to use for computation.

        Returns:
            Series: The computed column as a pandas Series.
        """
        return self.compute_func(df)


class MergeColumns(Transformation):
    """
    MergeColumns

    Merges multiple columns into a single column by concatenating their string
    representations using a specified separator.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        columns (list): The list of column names to merge.
        new_column (str): The name of the new merged column.
        separator (str): The string used to separate the merged values.
    """
    def __init__(self,
                 dataframe,
                 columns,
                 new_column,
                 separator=" ",
                 step_name="MergeColumns",
                 on_error=None):
        """
        Initializes the MergeColumns transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update.
            columns (list): A list of column names to merge.
            new_column (str): The name of the new column that will contain the merged values.
            separator (str, optional): The separator to use between merged column values.
                                       Defaults to a space.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "MergeColumns".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.columns = columns
        self.new_column = new_column
        self.separator = separator
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the MergeColumns transformation.

        Retrieves the DataFrame from the context, merges the specified columns by concatenating
        their string representations, adds the new merged column to the DataFrame, updates
        the context, and returns it.

        Arguments:
            context (Context): The context containing the DataFrame.

        Returns:
            Context: The updated context with the merged column.
        """
        df = context.dataframes[self.dataframe]
        df[self.new_column] = self.__merge_columns(df)
        context.set_dataframe(self.dataframe, df)
        return context

    def __merge_columns(self, df):
        """
        Merges the specified columns by concatenating their string representations.

        Arguments:
            df (DataFrame): The DataFrame containing the columns.

        Returns:
            Series: A Series with the merged column values.
        """
        return df[self.columns].astype(str).agg(self.separator.join, axis=1)


class SplitColumn(Transformation):
    """
    SplitColumn

    Splits a single column in a DataFrame into multiple new columns based on a delimiter.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        column (str): The name of the column to split.
        new_columns (list): The list of new column names to create.
        delimiter (str): The delimiter used to split the column's values.
    """
    def __init__(self,
                 dataframe,
                 column,
                 new_columns,
                 delimiter=" ",
                 step_name="SplitColumn",
                 on_error=None):
        """
        Initializes the SplitColumn transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update.
            column (str): The name of the column to split.
            new_columns (list): A list of new column names to assign to the split parts.
            delimiter (str, optional): The delimiter used to split the column values.
                                       Defaults to a space.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "SplitColumn".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.column = column
        self.new_columns = new_columns
        self.delimiter = delimiter
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the SplitColumn transformation.

        Retrieves the DataFrame from the context, splits the specified column using the
        delimiter, assigns the resulting parts to new columns, concatenates them with the
        original DataFrame, updates the context, and returns it.

        Arguments:
            context (Context): The context containing the DataFrame.

        Returns:
            Context: The updated context with the split columns added.
        """
        df = context.dataframes[self.dataframe]
        context.set_dataframe(self.dataframe, self.__split_column(df))
        return context

    def __split_column(self, df):
        """
        Splits the specified column into multiple columns.

        Arguments:
            df (DataFrame): The DataFrame containing the column.

        Returns:
            DataFrame: The DataFrame with the new split columns concatenated.
        """
        splits = df[self.column].str.split(self.delimiter, expand=True)
        splits.columns = self.new_columns
        df = pd.concat([df, splits], axis=1)
        return df


class ExplodeColumn(Transformation):
    """
    ExplodeColumn

    Explodes a column containing list-like elements into multiple rows, duplicating the other
    column values.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        column (str): The name of the column to explode.
    """
    def __init__(self,
                 dataframe,
                 column,
                 step_name="ExplodeColumn",
                 on_error=None):
        """
        Initializes the ExplodeColumn transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update.
            column (str): The name of the column to explode.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "ExplodeColumn".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.column = column
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)

    def func(self, context):
        """
        Executes the ExplodeColumn transformation.

        Retrieves the DataFrame from the context, explodes the specified column,
        updates the DataFrame in the context, and returns the updated context.

        Arguments:
            context (Context): The context containing the DataFrame.

        Returns:
            Context: The updated context with the exploded column.
        """
        df = context.dataframes[self.dataframe]
        context.set_dataframe(self.dataframe, self.__explode_column(df))
        return context

    def __explode_column(self, df):
        """
        Explodes the specified column into multiple rows.

        Arguments:
            df (DataFrame): The DataFrame containing the column.

        Returns:
            DataFrame: The DataFrame after exploding the column.
        """
        return df.explode(self.column)


class CreateColumnFromVariable(Transformation):
    """
    CreateColumnFromVariable

    Creates a new column in a DataFrame with a constant value derived from a variable.
    This transformation is used to add a new column whose values are all set to the provided
    variable's value.

    Attributes:
        dataframe (str): The name of the DataFrame in the context.
        column (str): The name of the new column to be created.
        variable: The constant value to assign to the new column.
    """
    def __init__(self,
                 dataframe,
                 column,
                 variable,
                 step_name="CreateColumnFromVariable",
                 on_error=None):
        """
        Initializes the CreateColumnFromVariable transformation.

        Arguments:
            dataframe (str): The name of the DataFrame to update.
            column (str): The name of the new column to create.
            variable: The constant value to assign to the new column.
            step_name (str, optional): The name of the transformation step.
                                       Defaults to "CreateColumnFromVariable".
            on_error (str, optional): The error handling strategy.
        """
        self.dataframe = dataframe
        self.column = column
        self.variable = variable
        super().__init__(step_name=step_name,
                         func=self.func,
                         dataframes=dataframe,
                         on_error=on_error)
        self.update_return_list(self.variable)
        self.update_params_list(self.variable)

    def func(self, context, **kwargs):
        """
        Executes the CreateColumnFromVariable transformation.

        Retrieves the DataFrame from the context, creates a new column with the constant
        value provided by the variable, updates the DataFrame in the context, and returns
        the modified context.

        Arguments:
            context (Context): The context containing the DataFrame.
            **kwargs: Keyword arguments that should include the variable value.

        Returns:
            Context: The updated context with the new column added.
        """
        df = context.dataframes[self.dataframe]
        df[self.column] = kwargs[self.variable]
        context.set_dataframe(self.dataframe, df)
        return context



"""
Module: types

This module provides a suite of type validation functions to ensure we conform to expected types.
These functions validate whether a given object qualifies as an:
extractor, multi-extractor, loader, step, or context.
Using these type-checking functions, Pipeline can enforce proper usage of its components.
"""
from .type_validation import is_extractor
from .type_validation import is_multiextractor
from .type_validation import is_loader
from .type_validation import is_step
from .type_validation import is_context
from .type_validation import is_context_object



"""
Module: direct_chunker

This module implements the DirectChunker class, an implementation of the Chunker abstract class.
DirectChunker computes chunk coordinates based on each step's defined chunk size and the
total number of rows. It enqueues tuples representing the start and stop indices for processing
each chunk, ensuring that data is divided into manageable segments. 
Once all chunks for a step are processed, it pads the coordinate queue with (None, None) tuples.
"""
from .chunker import Chunker

class DirectChunker(Chunker):
    """
    DirectChunker

    A concrete implementation of the Chunker class that calculates chunk coordinates directly.
    It iterates through the steps that support chunking, computes the start and stop indices 
    based on the chunk size and total row count, and populates the coordinate queue with values.
    If a step has finished processing all available rows, a (None, None) tuple is enqueued.
    """
    def __init__(self, step_index):
        """
        Direct Chunker Class Constructor method
        Initializes the DirectChunker object by invoking the parent class constructor and
        then invoking the calculate_chunks() method to populate the coordinate queue.

        Arguments:
            step_index (OrderedDict): 
                An ordered dictionary mapping step keys to step objects in the Pipeline.
        """
        super().__init__(step_index=step_index)

    def calculate_chunks(self):
        """
        Public method: calculate_chunks()
        Calculates the chunk coordinates for each step in the chunk index.
        The chunk coordinates are added to the coordinate queue.
        Calculates by using the chunk size and the number of rows in the step.
        If the chunk size is greater than the number of rows, the chunk size is set
        to the number of rows.
        A (None, None) tuple is added to the queue for each step that has finished calculating.

        This version produces tuples of (start_index, nrows) for compatibility with
        pandas read_csv's skiprows and nrows.
        """
        chunk_keys = list(self.chunk_index.keys())
        num_keys = len(chunk_keys)
        it = 0

        while any(not self.chunk_index[key][3] for key in chunk_keys):
            key = chunk_keys[it]
            it = (it + 1) % num_keys

            chunk_size, current_chunk, num_rows, finished_calculating = self.chunk_index[key]

            if finished_calculating:
                start_idx = None
                nrows = None
            else:
                start_idx = current_chunk * chunk_size
                if start_idx <= num_rows:
                    stop_idx = start_idx + chunk_size
                    if stop_idx >= num_rows:
                        stop_idx = num_rows
                        finished_calculating = True
                    nrows = stop_idx - start_idx
                else:
                    # If start_idx is beyond num_rows, there's nothing to read.
                    start_idx = None
                    nrows = None
                    finished_calculating = True

            self.coordinate_queue.put((start_idx, nrows))
            new_current_chunk = current_chunk + (1 if not finished_calculating else 0)
            self.chunk_index[key] = (chunk_size, new_current_chunk, num_rows, finished_calculating)

        # pad the final output so that the queue size is a multiple of the number of keys
        while self.coordinate_queue.qsize() % num_keys != 0:
            self.coordinate_queue.put((None, None))
